#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Tests for plane_wave_hamiltonian.py"""
import unittest

import numpy as np

from openfermion.hamiltonians.plane_wave_hamiltonian import (
    jellium_model, jordan_wigner_dual_basis_hamiltonian, plane_wave_hamiltonian)
from openfermion.transforms.opconversions import jordan_wigner
from openfermion.linalg import eigenspectrum, get_sparse_operator
from openfermion.utils import Grid, is_hermitian


class PlaneWaveHamiltonianTest(unittest.TestCase):

    def test_plane_wave_hamiltonian_integration(self):
        length_set = [2, 3, 4]
        spinless_set = [True, False]
        length_scale = 1.1
        for geometry in [[('H', (0,)), ('H', (0.8,))], [('H', (0.1,))],
                         [('H', (0.1,))]]:
            for l in length_set:
                for spinless in spinless_set:
                    grid = Grid(dimensions=1, scale=length_scale, length=l)
                    h_plane_wave = plane_wave_hamiltonian(
                        grid, geometry, spinless, True, include_constant=False)
                    h_dual_basis = plane_wave_hamiltonian(
                        grid, geometry, spinless, False, include_constant=False)

                    # Test for Hermiticity
                    plane_wave_operator = get_sparse_operator(h_plane_wave)
                    dual_operator = get_sparse_operator(h_dual_basis)
                    self.assertTrue(is_hermitian((plane_wave_operator)))
                    self.assertTrue(is_hermitian(dual_operator))

                    jw_h_plane_wave = jordan_wigner(h_plane_wave)
                    jw_h_dual_basis = jordan_wigner(h_dual_basis)
                    h_plane_wave_spectrum = eigenspectrum(jw_h_plane_wave)
                    h_dual_basis_spectrum = eigenspectrum(jw_h_dual_basis)

                    max_diff = np.amax(h_plane_wave_spectrum -
                                       h_dual_basis_spectrum)
                    min_diff = np.amin(h_plane_wave_spectrum -
                                       h_dual_basis_spectrum)
                    self.assertAlmostEqual(max_diff, 0)
                    self.assertAlmostEqual(min_diff, 0)

    def test_plane_wave_hamiltonian_default_to_jellium_with_no_geometry(self):
        grid = Grid(dimensions=1, scale=1.0, length=4)
        self.assertTrue(plane_wave_hamiltonian(grid) == jellium_model(grid))

    def test_plane_wave_hamiltonian_bad_geometry(self):
        grid = Grid(dimensions=1, scale=1.0, length=4)
        with self.assertRaises(ValueError):
            plane_wave_hamiltonian(grid, geometry=[('H', (0, 0, 0))])

        with self.assertRaises(ValueError):
            plane_wave_hamiltonian(grid,
                                   geometry=[('H', (0, 0, 0))],
                                   include_constant=True)

    def test_plane_wave_hamiltonian_bad_element(self):
        grid = Grid(dimensions=3, scale=1.0, length=4)
        with self.assertRaises(ValueError):
            plane_wave_hamiltonian(grid, geometry=[('Unobtainium', (0, 0, 0))])

    def test_jordan_wigner_dual_basis_hamiltonian(self):
        grid = Grid(dimensions=2, length=3, scale=1.)
        spinless_set = [True, False]
        geometry = [('H', (0, 0)), ('H', (0.5, 0.8))]
        for spinless in spinless_set:
            fermion_hamiltonian = plane_wave_hamiltonian(grid,
                                                         geometry,
                                                         spinless,
                                                         False,
                                                         include_constant=False)
            qubit_hamiltonian = jordan_wigner(fermion_hamiltonian)

            test_hamiltonian = jordan_wigner_dual_basis_hamiltonian(
                grid, geometry, spinless, include_constant=False)
            self.assertTrue(test_hamiltonian == qubit_hamiltonian)

    def test_jordan_wigner_dual_basis_hamiltonian_default_to_jellium(self):
        grid = Grid(dimensions=1, scale=1.0, length=4)
        self.assertTrue(
            jordan_wigner_dual_basis_hamiltonian(grid) == jordan_wigner(
                jellium_model(grid, plane_wave=False)))

    def test_jordan_wigner_dual_basis_hamiltonian_bad_geometry(self):
        grid = Grid(dimensions=1, scale=1.0, length=4)
        with self.assertRaises(ValueError):
            jordan_wigner_dual_basis_hamiltonian(grid,
                                                 geometry=[('H', (0, 0, 0))])

        with self.assertRaises(ValueError):
            jordan_wigner_dual_basis_hamiltonian(grid,
                                                 geometry=[('H', (0, 0, 0))],
                                                 include_constant=True)

    def test_jordan_wigner_dual_basis_hamiltonian_bad_element(self):
        grid = Grid(dimensions=3, scale=1.0, length=4)
        with self.assertRaises(ValueError):
            jordan_wigner_dual_basis_hamiltonian(grid,
                                                 geometry=[('Unobtainium',
                                                            (0, 0, 0))])

    def test_plane_wave_energy_cutoff(self):
        geometry = [('H', (0,)), ('H', (0.8,))]
        grid = Grid(dimensions=1, scale=1.1, length=5)
        e_cutoff = 50.0

        h_1 = plane_wave_hamiltonian(grid, geometry, True, True, False)
        jw_1 = jordan_wigner(h_1)
        spectrum_1 = eigenspectrum(jw_1)

        h_2 = plane_wave_hamiltonian(grid, geometry, True, True, False,
                                     e_cutoff)
        jw_2 = jordan_wigner(h_2)
        spectrum_2 = eigenspectrum(jw_2)

        max_diff = np.amax(np.absolute(spectrum_1 - spectrum_2))
        self.assertGreater(max_diff, 0.)

    def test_plane_wave_period_cutoff(self):
        # TODO: After figuring out the correct formula for period cutoff for
        #     dual basis, change period_cutoff to default, and change
        #     h_1 to also accept period_cutoff for real integration test.

        geometry = [('H', (0,)), ('H', (0.8,))]
        grid = Grid(dimensions=1, scale=1.1, length=5)
        period_cutoff = 50.0

        h_1 = plane_wave_hamiltonian(grid, geometry, True, True, False, None)
        jw_1 = jordan_wigner(h_1)
        spectrum_1 = eigenspectrum(jw_1)

        h_2 = plane_wave_hamiltonian(grid, geometry, True, True, False, None,
                                     True, period_cutoff)
        jw_2 = jordan_wigner(h_2)
        spectrum_2 = eigenspectrum(jw_2)

        max_diff = np.amax(np.absolute(spectrum_1 - spectrum_2))
        self.assertGreater(max_diff, 0.)

        # TODO: This is only for code coverage. Remove after having real
        #     integration test.
        plane_wave_hamiltonian(grid, geometry, True, True, False, None, True)



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""tests for whether operators are contextual contextuality"""
from typing import List

from openfermion.ops.operators import QubitOperator


def _commutes(operator1: QubitOperator, operator2: QubitOperator) -> bool:
    return operator1 * operator2 == operator2 * operator1


def _non_fully_commuting_terms(hamiltonian: QubitOperator
                              ) -> List[QubitOperator]:
    terms = list([QubitOperator(key) for key in hamiltonian.terms.keys()])
    T = []  # will contain the subset of terms that do not
    # commute universally in terms
    for i in range(len(terms)):
        if any(not _commutes(terms[i], terms[j]) for j in range(len(terms))):
            T.append(terms[i])
    return T


def is_contextual(hamiltonian: QubitOperator) -> bool:
    """
    Determine whether a hamiltonian (instance of QubitOperator) is contextual,
    in the sense of https://arxiv.org/abs/1904.02260.

    Args:
        hamiltonian (QubitOperator): the hamiltonian whose
            contextuality is to be evaluated.

    Returns:
        boolean indicating whether hamiltonian is contextual or not.
    """
    if not isinstance(hamiltonian, QubitOperator):
        raise TypeError('Only supported for QubitOperators.')
    T = _non_fully_commuting_terms(hamiltonian)
    # Search in T for triples in which exactly one pair anticommutes;
    # if any exist, hamiltonian is contextual.
    for i in range(len(T)):  # WLOG, i indexes the operator that (putatively)
        # commutes with both others.
        for j in range(len(T)):
            for k in range(j + 1, len(T)):  # Ordering of j, k does not matter.
                if i!=j and i!=k and _commutes(T[i],T[j]) \
                    and _commutes(T[i],T[k]) and \
                        not _commutes(T[j],T[k]):
                    return True
    return False



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""A Trotter algorithm using the low rank decomposition strategy."""

from typing import cast, Optional, Sequence, TYPE_CHECKING, Tuple

import numpy

import cirq

import openfermion.ops as ops
# import openfermion.circuits.gates as gates
from openfermion.circuits.gates import rot11, rot111
# import openfermion.circuits.primitives as primitives
from openfermion.circuits.primitives import swap_network, bogoliubov_transform
from openfermion.circuits.low_rank import (low_rank_two_body_decomposition,
                                           prepare_one_body_squared_evolution)
from openfermion.circuits.trotter.trotter_algorithm import (Hamiltonian,
                                                            TrotterStep,
                                                            TrotterAlgorithm)

if TYPE_CHECKING:
    # pylint: disable=unused-import
    from typing import List


class LowRankTrotterAlgorithm(TrotterAlgorithm):
    r"""A Trotter algorithm using the low rank decomposition strategy.

    This algorithm simulates an InteractionOperator with real coefficients.
    The one-body terms are simulated in their diagonal basis; the basis change
    is accomplished with a Bogoliubov transformation.
    To simulate the two-body terms, the two-body tensor is decomposed into
    singular components and possibly truncating. Then, each singular component
    is simulated in the appropriate basis using a (non-fermionic) swap network.
    The general idea is based on expressing the two-body operator as
    :math:`\sum_{pqrs} h_{pqrs} a^\dagger_p a^\dagger_q a_r a_s =
    \sum_{j=0}^{J-1} \lambda_j (\sum_{pq} g_{jpq} a^\dagger_p a_q)^2`
    One can then diagonalize the squared one-body component as
    math:`\sum_{pq} g_{pqj} a^\dagger_p a_q =
    R_j (\sum_{p} f_{pj} n_p) R_j^\dagger`
    Then, a 'low rank' Trotter step of the two-body tensor can be simulated as
    :math:`\prod_{j=0}^{J-1}
    R_j e^{-i \lambda_j \sum_{pq} f_{pj} f_{qj} n_p n_q} R_j^\dagger`.
    The :math:`R_j` are Bogoliubov transformations, and one can
    use a swap network to simulate the diagonal :math:`n_p n_q` terms.
    The value of J is either fully the square of the number of qubits,
    which would imply no truncation, or it is specified by the user,
    or it is chosen so that
    :math:`\sum_{l=0}^{L-1} (\sum_{pq} |g_{lpq}|)^2 |\lambda_l| < x`
    where x is a truncation threshold specified by user.
    """

    supported_types = {ops.InteractionOperator}

    def __init__(self,
                 truncation_threshold: Optional[float] = 1e-8,
                 final_rank: Optional[int] = None,
                 spin_basis=True) -> None:
        """
        Args:
            truncation_threshold: The value of x from the docstring of
                this class.
            final_rank: If provided, this specifies the value of J at which to
                truncate.
            spin_basis: Whether the Hamiltonian is given in the spin orbital
                (rather than spatial orbital) basis.
        """
        self.truncation_threshold = truncation_threshold
        self.final_rank = final_rank
        self.spin_basis = spin_basis

    def asymmetric(self, hamiltonian: Hamiltonian) -> Optional[TrotterStep]:
        return AsymmetricLowRankTrotterStep(hamiltonian,
                                            self.truncation_threshold,
                                            self.final_rank, self.spin_basis)

    def controlled_asymmetric(self, hamiltonian: Hamiltonian
                             ) -> Optional[TrotterStep]:
        return ControlledAsymmetricLowRankTrotterStep(hamiltonian,
                                                      self.truncation_threshold,
                                                      self.final_rank,
                                                      self.spin_basis)


LOW_RANK = LowRankTrotterAlgorithm()


class LowRankTrotterStep(TrotterStep):

    def __init__(self,
                 hamiltonian: 'ops.InteractionOperator',
                 truncation_threshold: Optional[float] = 1e-8,
                 final_rank: Optional[int] = None,
                 spin_basis=True) -> None:

        self.truncation_threshold = truncation_threshold
        self.final_rank = final_rank

        # Perform the low rank decomposition of two-body operator.
        self.eigenvalues, self.one_body_squares, one_body_correction, _ = (
            low_rank_two_body_decomposition(
                hamiltonian.two_body_tensor,
                truncation_threshold=self.truncation_threshold,
                final_rank=self.final_rank,
                spin_basis=spin_basis))

        # Get scaled density-density terms and basis transformation matrices.
        self.scaled_density_density_matrices = []  # type: List[numpy.ndarray]
        self.basis_change_matrices = []  # type: List[numpy.ndarray]
        for j in range(len(self.eigenvalues)):
            density_density_matrix, basis_change_matrix = (
                prepare_one_body_squared_evolution(self.one_body_squares[j]))
            self.scaled_density_density_matrices.append(
                numpy.real(self.eigenvalues[j] * density_density_matrix))
            self.basis_change_matrices.append(basis_change_matrix)

        # Get transformation matrix and orbital energies for one-body terms
        one_body_coefficients = (hamiltonian.one_body_tensor +
                                 one_body_correction)
        quad_ham = ops.QuadraticHamiltonian(one_body_coefficients)
        self.one_body_energies, self.one_body_basis_change_matrix, _ = (
            quad_ham.diagonalizing_bogoliubov_transform())

        super().__init__(hamiltonian)


class AsymmetricLowRankTrotterStep(LowRankTrotterStep):

    def trotter_step(self,
                     qubits: Sequence[cirq.Qid],
                     time: float,
                     control_qubit: Optional[cirq.Qid] = None) -> cirq.OP_TREE:

        n_qubits = len(qubits)

        # Change to the basis in which the one-body term is diagonal
        yield bogoliubov_transform(qubits,
                                   self.one_body_basis_change_matrix.T.conj())

        # Simulate the one-body terms.
        for p in range(n_qubits):
            yield cirq.rz(rads=-self.one_body_energies[p] * time).on(qubits[p])

        # Simulate each singular vector of the two-body terms.
        prior_basis_matrix = self.one_body_basis_change_matrix

        for j in range(len(self.eigenvalues)):

            # Get the two-body coefficients and basis change matrix.
            two_body_coefficients = self.scaled_density_density_matrices[j]
            basis_change_matrix = self.basis_change_matrices[j]

            # Merge previous basis change matrix with the inverse of the
            # current one
            merged_basis_change_matrix = numpy.dot(prior_basis_matrix,
                                                   basis_change_matrix.T.conj())
            yield bogoliubov_transform(qubits, merged_basis_change_matrix)

            # Simulate the off-diagonal two-body terms.
            yield swap_network(
                qubits, lambda p, q, a, b: rot11(
                    rads=-2 * two_body_coefficients[p, q] * time).on(a, b))
            qubits = qubits[::-1]

            # Simulate the diagonal two-body terms.
            for p in range(n_qubits):
                yield cirq.rz(rads=-two_body_coefficients[p, p] * time).on(
                    qubits[p])

            # Update prior basis change matrix
            prior_basis_matrix = basis_change_matrix

        # Undo final basis transformation
        yield bogoliubov_transform(qubits, prior_basis_matrix)

    def step_qubit_permutation(
            self,
            qubits: Sequence[cirq.Qid],
            control_qubit: Optional[cirq.Qid] = None
    ) -> Tuple[Sequence[cirq.Qid], Optional[cirq.Qid]]:
        # A Trotter step reverses the qubit ordering when the number of
        # eigenvalues is odd
        if len(self.eigenvalues) & 1:
            return qubits[::-1], None
        else:
            return qubits, None

    def finish(self,
               qubits: Sequence[cirq.Qid],
               n_steps: int,
               control_qubit: Optional[cirq.Qid] = None,
               omit_final_swaps: bool = False) -> cirq.OP_TREE:
        if not omit_final_swaps:
            # If the number of swap networks was odd, swap the qubits back
            if n_steps & 1 and len(self.eigenvalues) & 1:
                yield swap_network(qubits)


class ControlledAsymmetricLowRankTrotterStep(LowRankTrotterStep):

    def trotter_step(self,
                     qubits: Sequence[cirq.Qid],
                     time: float,
                     control_qubit: Optional[cirq.Qid] = None) -> cirq.OP_TREE:

        if not isinstance(control_qubit, cirq.Qid):
            raise TypeError('Control qudit must be specified.')

        n_qubits = len(qubits)

        # Change to the basis in which the one-body term is diagonal
        yield bogoliubov_transform(qubits,
                                   self.one_body_basis_change_matrix.T.conj())

        # Simulate the one-body terms.
        for p in range(n_qubits):
            yield rot11(rads=-self.one_body_energies[p] * time).on(
                control_qubit, qubits[p])

        # Simulate each singular vector of the two-body terms.
        prior_basis_matrix = self.one_body_basis_change_matrix

        for j in range(len(self.eigenvalues)):

            # Get the two-body coefficients and basis change matrix.
            two_body_coefficients = self.scaled_density_density_matrices[j]
            basis_change_matrix = self.basis_change_matrices[j]

            # Merge previous basis change matrix with the inverse of the
            # current one
            merged_basis_change_matrix = numpy.dot(prior_basis_matrix,
                                                   basis_change_matrix.T.conj())
            yield bogoliubov_transform(qubits, merged_basis_change_matrix)

            # Simulate the off-diagonal two-body terms.
            yield swap_network(
                qubits, lambda p, q, a, b: rot111(-2 * two_body_coefficients[
                    p, q] * time).on(cast(cirq.Qid, control_qubit), a, b))
            qubits = qubits[::-1]

            # Simulate the diagonal two-body terms.
            yield (rot11(rads=-two_body_coefficients[k, k] * time).on(
                control_qubit, qubits[k]) for k in range(n_qubits))

            # Update prior basis change matrix.
            prior_basis_matrix = basis_change_matrix

        # Undo final basis transformation.
        yield bogoliubov_transform(qubits, prior_basis_matrix)

        # Apply phase from constant term
        yield cirq.rz(rads=-self.hamiltonian.constant * time).on(control_qubit)

    def step_qubit_permutation(
            self,
            qubits: Sequence[cirq.Qid],
            control_qubit: Optional[cirq.Qid] = None
    ) -> Tuple[Sequence[cirq.Qid], Optional[cirq.Qid]]:
        # A Trotter step reverses the qubit ordering when the number of
        # eigenvalues is odd
        if len(self.eigenvalues) & 1:
            return qubits[::-1], control_qubit
        else:
            return qubits, control_qubit

    def finish(self,
               qubits: Sequence[cirq.Qid],
               n_steps: int,
               control_qubit: Optional[cirq.Qid] = None,
               omit_final_swaps: bool = False) -> cirq.OP_TREE:
        if not omit_final_swaps:
            # If the number of swap networks was odd, swap the qubits back
            if n_steps & 1 and len(self.eigenvalues) & 1:
                yield swap_network(qubits)



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Operations to remove qubits/spin orbitals from operators"""

import copy

from openfermion.ops.operators import (BosonOperator, FermionOperator,
                                       MajoranaOperator, QuadOperator,
                                       QubitOperator)


def freeze_orbitals(fermion_operator, occupied, unoccupied=None, prune=True):
    """Fix some orbitals to be occupied and others unoccupied.

    Removes all operators acting on the specified orbitals, and renumbers the
    remaining orbitals to eliminate unused indices. The sign of each term
    is modified according to the ladder uperator anti-commutation relations in
    order to preserve the expectation value of the operator.

    Args:
        occupied: A list containing the indices of the orbitals that are to be
            assumed to be occupied.
        unoccupied: A list containing the indices of the orbitals that are to
            be assumed to be unoccupied.
    """
    new_operator = fermion_operator
    frozen = [(index, 1) for index in occupied]
    if unoccupied is not None:
        frozen += [(index, 0) for index in unoccupied]

    # Loop over each orbital to be frozen. Within each term, move all
    # ops acting on that orbital to the right side of the term, keeping
    # track of sign flips that come from swapping operators.
    for item in frozen:
        tmp_operator = FermionOperator()
        for term in new_operator.terms:
            new_term = []
            new_coef = new_operator.terms[term]
            current_occupancy = item[1]
            n_ops = 0  # Number of operations on index that have been moved
            n_swaps = 0  # Number of swaps that have been done

            for op in enumerate(reversed(term)):
                if op[1][0] is item[0]:
                    n_ops += 1

                    # Determine number of swaps needed to bring the op in
                    # front of all ops acting on other indices
                    n_swaps += op[0] - n_ops

                    # Check if the op annihilates the current state
                    if current_occupancy == op[1][1]:
                        new_coef = 0

                    # Update current state
                    current_occupancy = (current_occupancy + 1) % 2
                else:
                    new_term.insert(0, op[1])
            if n_swaps % 2:
                new_coef *= -1
            if new_coef and current_occupancy == item[1]:
                tmp_operator += FermionOperator(tuple(new_term), new_coef)
        new_operator = tmp_operator

    # For occupied frozen orbitals, we must also bring together the creation
    # operator from the ket and the annihilation operator from the bra when
    # evaluating expectation values. This can result in an additional minus
    # sign.
    for term in new_operator.terms:
        for index in occupied:
            for op in term:
                if op[0] > index:
                    new_operator.terms[term] *= -1

    # Renumber indices to remove frozen orbitals
    new_operator = prune_unused_indices(new_operator)

    return new_operator


def prune_unused_indices(symbolic_operator):
    """
    Remove indices that do not appear in any terms.

    Indices will be renumbered such that if an index i does not appear in
    any terms, then the next largest index that appears in at least one
    term will be renumbered to i.
    """

    # Determine which indices appear in at least one term
    indices = []
    for term in symbolic_operator.terms:
        for op in term:
            if op[0] not in indices:
                indices.append(op[0])
    indices.sort()

    # Construct a dict that maps the old indices to new ones
    index_map = {}
    for index in enumerate(indices):
        index_map[index[1]] = index[0]

    new_operator = copy.deepcopy(symbolic_operator)
    new_operator.terms.clear()

    # Replace the indices in the terms with the new indices
    for term in symbolic_operator.terms:
        new_term = [(index_map[op[0]], op[1]) for op in term]
        new_operator.terms[tuple(new_term)] = symbolic_operator.terms[term]

    return new_operator



"""Tests for jellium_hf_state.py."""

from itertools import permutations
import unittest
import numpy

from openfermion.hamiltonians import jellium_model, wigner_seitz_length_scale
from openfermion.linalg import (get_sparse_operator, get_ground_state,
                                expectation, jw_number_restrict_operator)
from openfermion.utils.grid import Grid
from openfermion.hamiltonians.jellium_hf_state import hartree_fock_state_jellium


class JelliumHartreeFockStateTest(unittest.TestCase):

    def test_hf_state_energy_close_to_ground_energy_at_high_density(self):
        grid_length = 8
        dimension = 1
        spinless = True
        n_particles = grid_length**dimension // 2

        # High density -> small length_scale.
        length_scale = 0.25

        grid = Grid(dimension, grid_length, length_scale)
        hamiltonian = jellium_model(grid, spinless)
        hamiltonian_sparse = get_sparse_operator(hamiltonian)

        hf_state = hartree_fock_state_jellium(grid,
                                              n_particles,
                                              spinless,
                                              plane_wave=True)

        restricted_hamiltonian = jw_number_restrict_operator(
            hamiltonian_sparse, n_particles)

        E_g = get_ground_state(restricted_hamiltonian)[0]
        E_HF_plane_wave = expectation(hamiltonian_sparse, hf_state)

        self.assertAlmostEqual(E_g, E_HF_plane_wave, places=5)

    def test_hf_state_energy_same_in_plane_wave_and_dual_basis(self):
        grid_length = 4
        dimension = 1
        wigner_seitz_radius = 10.0
        spinless = False

        n_orbitals = grid_length**dimension
        if not spinless:
            n_orbitals *= 2
        n_particles = n_orbitals // 2

        length_scale = wigner_seitz_length_scale(wigner_seitz_radius,
                                                 n_particles, dimension)

        grid = Grid(dimension, grid_length, length_scale)
        hamiltonian = jellium_model(grid, spinless)
        hamiltonian_dual_basis = jellium_model(grid, spinless, plane_wave=False)

        # Get the Hamiltonians as sparse operators.
        hamiltonian_sparse = get_sparse_operator(hamiltonian)
        hamiltonian_dual_sparse = get_sparse_operator(hamiltonian_dual_basis)

        hf_state = hartree_fock_state_jellium(grid,
                                              n_particles,
                                              spinless,
                                              plane_wave=True)
        hf_state_dual = hartree_fock_state_jellium(grid,
                                                   n_particles,
                                                   spinless,
                                                   plane_wave=False)

        E_HF_plane_wave = expectation(hamiltonian_sparse, hf_state)
        E_HF_dual = expectation(hamiltonian_dual_sparse, hf_state_dual)

        self.assertAlmostEqual(E_HF_dual, E_HF_plane_wave)

    def test_hf_state_plane_wave_basis_lowest_single_determinant_state(self):
        grid_length = 7
        dimension = 1
        spinless = True
        n_particles = 4
        length_scale = 2.0

        grid = Grid(dimension, grid_length, length_scale)
        hamiltonian = jellium_model(grid, spinless)
        hamiltonian_sparse = get_sparse_operator(hamiltonian)

        hf_state = hartree_fock_state_jellium(grid,
                                              n_particles,
                                              spinless,
                                              plane_wave=True)

        HF_energy = expectation(hamiltonian_sparse, hf_state)

        for occupied_orbitals in permutations([1] * n_particles + [0] *
                                              (grid_length - n_particles)):
            state_index = numpy.sum(2**numpy.array(occupied_orbitals))
            HF_competitor = numpy.zeros(2**grid_length)
            HF_competitor[state_index] = 1.0

            self.assertLessEqual(HF_energy,
                                 expectation(hamiltonian_sparse, HF_competitor))



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import itertools
import fractions
import unittest

import numpy

from openfermion.ops.operators import QubitOperator
from openfermion.transforms.opconversions import get_fermion_operator
from openfermion.utils import count_qubits, is_hermitian
from openfermion.testing.testing_utils import (
    EqualsTester, haar_random_vector, random_antisymmetric_matrix,
    random_diagonal_coulomb_hamiltonian, random_hermitian_matrix,
    random_interaction_operator, random_quadratic_hamiltonian,
    random_qubit_operator, random_unitary_matrix, module_importable,
    _ClassUnknownToSubjects)


def test_random_qubit_operator():
    op = random_qubit_operator(n_qubits=20,
                               max_num_terms=20,
                               max_many_body_order=20)

    assert isinstance(op, QubitOperator)
    assert op.many_body_order() <= 20
    assert len(op.terms) <= 20
    assert count_qubits(op) <= 20


def test_module_importable():
    assert module_importable('numpy') is True
    assert module_importable('some_random_crap') is False


def test_hashing_unknown_class():
    unknown_object = _ClassUnknownToSubjects()
    test_dict = {unknown_object: 0}
    assert test_dict[unknown_object] == 0


class EqualsTesterTest(unittest.TestCase):

    def test_add_equality_group_correct(self):
        eq = EqualsTester(self)

        eq.add_equality_group(fractions.Fraction(1, 1))

        eq.add_equality_group(fractions.Fraction(1, 2),
                              fractions.Fraction(2, 4))

        eq.add_equality_group(fractions.Fraction(2, 3),
                              fractions.Fraction(12, 18),
                              fractions.Fraction(14, 21))

        eq.add_equality_group(2, 2.0, fractions.Fraction(2, 1))

        eq.add_equality_group([1, 2, 3], [1, 2, 3])

        eq.add_equality_group({'b': 3, 'a': 2}, {'a': 2, 'b': 3})

        eq.add_equality_group('unrelated')

    def test_assert_add_equality_pair(self):
        eq = EqualsTester(self)

        with self.assertRaises(AssertionError):
            eq.make_equality_pair(object)

        eq.make_equality_pair(lambda: 1)
        eq.make_equality_pair(lambda: 2)
        eq.add_equality_group(3)

        with self.assertRaises(AssertionError):
            eq.add_equality_group(1)
        with self.assertRaises(AssertionError):
            eq.make_equality_pair(lambda: 1)
        with self.assertRaises(AssertionError):
            eq.make_equality_pair(lambda: 3)

    def test_add_equality_group_not_equivalent(self):
        eq = EqualsTester(self)
        with self.assertRaises(AssertionError):
            eq.add_equality_group(1, 2)

    def test_add_equality_group_not_disjoint(self):
        eq = EqualsTester(self)
        eq.add_equality_group(1)
        with self.assertRaises(AssertionError):
            eq.add_equality_group(1)

    def test_add_equality_group_bad_hash(self):

        class KeyHash(object):

            def __init__(self, k, h):
                self._k = k
                self._h = h

            def __eq__(self, other):
                return isinstance(other, KeyHash) and self._k == other._k

            def __ne__(self, other):
                return not self == other

            def __hash__(self):
                return self._h

        eq = EqualsTester(self)
        eq.add_equality_group(KeyHash('a', 5), KeyHash('a', 5))
        eq.add_equality_group(KeyHash('b', 5))
        with self.assertRaises(AssertionError):
            eq.add_equality_group(KeyHash('c', 2), KeyHash('c', 3))

    def test_add_equality_group_exception_hash(self):

        class FailHash(object):

            def __hash__(self):
                raise ValueError('injected failure')

        eq = EqualsTester(self)
        with self.assertRaises(ValueError):
            eq.add_equality_group(FailHash())

    def test_can_fail_when_forgot_type_check(self):
        eq = EqualsTester(self)

        class NoTypeCheckEqualImplementation(object):

            def __init__(self):
                self.x = 1

            def __eq__(self, other):
                return self.x == other.x

            def __ne__(self, other):
                return not self == other

        with self.assertRaises(AttributeError):
            eq.add_equality_group(NoTypeCheckEqualImplementation())

    def test_fails_hash_is_default_and_inconsistent(self):
        eq = EqualsTester(self)

        class DefaultHashImplementation(object):
            __hash__ = object.__hash__

            def __init__(self):
                self.x = 1

            def __eq__(self, other):
                if not isinstance(other, type(self)):
                    return NotImplemented
                return self.x == other.x

            def __ne__(self, other):
                return not self == other

        with self.assertRaises(AssertionError):
            eq.make_equality_pair(DefaultHashImplementation)

    def test_fails_when_ne_is_inconsistent(self):
        eq = EqualsTester(self)

        class InconsistentNeImplementation(object):

            def __init__(self):
                self.x = 1

            def __eq__(self, other):
                return self.x == other.x

            def __ne__(self, other):
                return NotImplemented

        with self.assertRaises(AssertionError):
            eq.make_equality_pair(InconsistentNeImplementation)

    def test_fails_when_not_reflexive(self):
        eq = EqualsTester(self)

        class NotReflexiveImplementation(object):

            def __init__(self):
                self.x = 1

            def __eq__(self, other):
                return False

        with self.assertRaises(AssertionError):
            eq.add_equality_group(NotReflexiveImplementation())

    def test_fails_when_not_commutative(self):
        eq = EqualsTester(self)

        class NotCommutativeImplementation(object):

            def __init__(self, x):
                self.x = x

            def __eq__(self, other):
                return self.x <= other.x

            def __ne__(self, other):
                return not self == other

        with self.assertRaises(AssertionError):
            eq.add_equality_group(NotCommutativeImplementation(0),
                                  NotCommutativeImplementation(1))

        with self.assertRaises(AssertionError):
            eq.add_equality_group(NotCommutativeImplementation(1),
                                  NotCommutativeImplementation(0))


class RandomInteractionOperatorTest(unittest.TestCase):

    def test_hermiticity(self):
        n_orbitals = 5

        # Real, no spin
        iop = random_interaction_operator(n_orbitals, real=True)
        ferm_op = get_fermion_operator(iop)
        self.assertTrue(is_hermitian(ferm_op))

        # Real, spin
        iop = random_interaction_operator(n_orbitals,
                                          expand_spin=True,
                                          real=True)
        ferm_op = get_fermion_operator(iop)
        self.assertTrue(is_hermitian(ferm_op))

        # Complex, no spin
        iop = random_interaction_operator(n_orbitals, real=False)
        ferm_op = get_fermion_operator(iop)
        self.assertTrue(is_hermitian(ferm_op))

        # Complex, spin
        iop = random_interaction_operator(n_orbitals,
                                          expand_spin=True,
                                          real=False)
        ferm_op = get_fermion_operator(iop)
        self.assertTrue(is_hermitian(ferm_op))

    def test_symmetry(self):
        n_orbitals = 5

        # Real.
        iop = random_interaction_operator(n_orbitals,
                                          expand_spin=False,
                                          real=True)
        ferm_op = get_fermion_operator(iop)
        self.assertTrue(is_hermitian(ferm_op))
        two_body_coefficients = iop.two_body_tensor
        for p, q, r, s in itertools.product(range(n_orbitals), repeat=4):

            self.assertAlmostEqual(two_body_coefficients[p, q, r, s],
                                   two_body_coefficients[r, q, p, s])

            self.assertAlmostEqual(two_body_coefficients[p, q, r, s],
                                   two_body_coefficients[p, s, r, q])

            self.assertAlmostEqual(two_body_coefficients[p, q, r, s],
                                   two_body_coefficients[s, r, q, p])

            self.assertAlmostEqual(two_body_coefficients[p, q, r, s],
                                   two_body_coefficients[q, p, s, r])

            self.assertAlmostEqual(two_body_coefficients[p, q, r, s],
                                   two_body_coefficients[r, s, p, q])

            self.assertAlmostEqual(two_body_coefficients[p, q, r, s],
                                   two_body_coefficients[s, p, q, r])

            self.assertAlmostEqual(two_body_coefficients[p, q, r, s],
                                   two_body_coefficients[q, r, s, p])


class HaarRandomVectorTest(unittest.TestCase):

    def test_vector_norm(self):
        n = 15
        seed = 8317
        vector = haar_random_vector(n, seed)
        norm = vector.dot(numpy.conjugate(vector))
        self.assertAlmostEqual(1. + 0.j, norm)


class RandomSeedingTest(unittest.TestCase):

    def test_random_operators_are_reproducible(self):
        op1 = random_diagonal_coulomb_hamiltonian(5, seed=5947)
        op2 = random_diagonal_coulomb_hamiltonian(5, seed=5947)
        numpy.testing.assert_allclose(op1.one_body, op2.one_body)
        numpy.testing.assert_allclose(op1.two_body, op2.two_body)

        op1 = random_interaction_operator(5, seed=8911)
        op2 = random_interaction_operator(5, seed=8911)
        numpy.testing.assert_allclose(op1.one_body_tensor, op2.one_body_tensor)
        numpy.testing.assert_allclose(op1.two_body_tensor, op2.two_body_tensor)

        op1 = random_quadratic_hamiltonian(5, seed=17711)
        op2 = random_quadratic_hamiltonian(5, seed=17711)
        numpy.testing.assert_allclose(op1.combined_hermitian_part,
                                      op2.combined_hermitian_part)
        numpy.testing.assert_allclose(op1.antisymmetric_part,
                                      op2.antisymmetric_part)

        op1 = random_antisymmetric_matrix(5, seed=24074)
        op2 = random_antisymmetric_matrix(5, seed=24074)
        numpy.testing.assert_allclose(op1, op2)

        op1 = random_hermitian_matrix(5, seed=56753)
        op2 = random_hermitian_matrix(5, seed=56753)
        numpy.testing.assert_allclose(op1, op2)

        op1 = random_unitary_matrix(5, seed=56486)
        op2 = random_unitary_matrix(5, seed=56486)
        numpy.testing.assert_allclose(op1, op2)



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Code for evaluating Trotter errors for diagonal Coulomb Hamiltonians."""

import numpy

from openfermion.ops.operators import FermionOperator
from openfermion.transforms.opconversions import (get_fermion_operator,
                                                  normal_ordered)
from openfermion.utils.operator_utils import count_qubits
from openfermion.circuits.trotter.low_depth_trotter_error import (
    simulation_ordered_grouped_low_depth_terms_with_info)
from openfermion.transforms.opconversions import \
    commutator_ordered_diagonal_coulomb_with_two_body_operator


def diagonal_coulomb_potential_and_kinetic_terms_as_arrays(hamiltonian):
    """Give the potential and kinetic terms of a diagonal Coulomb Hamiltonian
    as arrays.

    Args:
        hamiltonian (FermionOperator): The diagonal Coulomb Hamiltonian to
                                       separate the potential and kinetic terms
                                       for. Identity is arbitrarily chosen
                                       to be part of the potential.

    Returns:
        Tuple of (potential_terms, kinetic_terms). Both elements of the tuple
        are numpy arrays of FermionOperators.
    """
    if not isinstance(hamiltonian, FermionOperator):
        try:
            hamiltonian = normal_ordered(get_fermion_operator(hamiltonian))
        except TypeError:
            raise TypeError('hamiltonian must be either a FermionOperator '
                            'or DiagonalCoulombHamiltonian.')

    potential = FermionOperator.zero()
    kinetic = FermionOperator.zero()

    for term, coeff in hamiltonian.terms.items():
        acted = set(term[i][0] for i in range(len(term)))
        if len(acted) == len(term) / 2:
            potential += FermionOperator(term, coeff)
        else:
            kinetic += FermionOperator(term, coeff)

    potential_terms = numpy.array([
        FermionOperator(term, coeff) for term, coeff in potential.terms.items()
    ])

    kinetic_terms = numpy.array(
        [FermionOperator(term, coeff) for term, coeff in kinetic.terms.items()])

    return (potential_terms, kinetic_terms)


def bit_mask_of_modes_acted_on_by_fermionic_terms(fermion_term_list,
                                                  n_qubits=None):
    """Create a mask of which modes of the system are acted on by which terms.

    Args:
        fermion_term_list (list of FermionOperators): A list of fermionic terms
            to calculate the bitmask for.
        n_qubits (int): The number of qubits (modes) in the system. If not
                        specified, defaults to the maximum of any term in
                        fermion_term_list.

    Returns:
        An n_qubits x len(fermion_term_list) boolean numpy array of whether
        each term acts on the given mode index.

    Raises:
        ValueError: if n_qubits is too small for the given terms.
    """
    if n_qubits is None:
        n_qubits = 0
        for term in fermion_term_list:
            n_qubits = max(n_qubits, count_qubits(term))

    mask = numpy.zeros((n_qubits, len(fermion_term_list)), dtype=bool)

    for term_number, term in enumerate(fermion_term_list):
        actions = term.terms
        for action in actions:
            for single_operator in action:
                mode = single_operator[0]
                try:
                    mask[mode][term_number] = True
                except IndexError:
                    raise ValueError('Bad n_qubits: must be greater than '
                                     'highest mode in any FermionOperator.')

    return mask


def split_operator_trotter_error_operator_diagonal_two_body(hamiltonian, order):
    """Compute the split-operator Trotter error of a diagonal two-body
    Hamiltonian.

    Args:
        hamiltonian (FermionOperator): The diagonal Coulomb Hamiltonian to
                                       compute the Trotter error for.
        order (str): Whether to simulate the split-operator Trotter step
                     with the kinetic energy T first (order='T+V') or with
                     the potential energy V first (order='V+T').

    Returns:
        error_operator: The second-order Trotter error operator.

    Notes:
        The second-order split-operator Trotter error is calculated from the
        double commutator [T, [V, T]] + [V, [V, T]] / 2 when T is simulated
        before V (i.e. exp(-iTt/2) exp(-iVt) exp(-iTt/2)), and from the
        double commutator [V, [T, V]] + [T, [T, V]] / 2 when V is simulated
        before T, following Equation 9 of "The Trotter Step Size Required for
        Accurate Quantum Simulation of Quantum Chemistry" by Poulin et al.
        The Trotter error operator is then obtained by dividing by 12.
    """
    n_qubits = count_qubits(hamiltonian)

    potential_terms, kinetic_terms = (
        diagonal_coulomb_potential_and_kinetic_terms_as_arrays(hamiltonian))

    # Cache halved potential and kinetic terms for the second commutator.
    halved_potential_terms = potential_terms / 2.0
    halved_kinetic_terms = kinetic_terms / 2.0

    # Assign the outer term of the second commutator based on the ordering.
    outer_potential_terms = (halved_potential_terms
                             if order == 'T+V' else potential_terms)
    outer_kinetic_terms = (halved_kinetic_terms
                           if order == 'V+T' else kinetic_terms)

    potential_mask = bit_mask_of_modes_acted_on_by_fermionic_terms(
        potential_terms, n_qubits)
    kinetic_mask = bit_mask_of_modes_acted_on_by_fermionic_terms(
        kinetic_terms, n_qubits)

    error_operator = FermionOperator.zero()

    for potential_term in potential_terms:
        modes_acted_on_by_potential_term = set()

        for potential_term_action in potential_term.terms:
            modes_acted_on_by_potential_term.update(
                set(operator[0] for operator in potential_term_action))

        if not modes_acted_on_by_potential_term:
            continue

        potential_term_mode_mask = numpy.logical_or.reduce(
            [kinetic_mask[mode] for mode in modes_acted_on_by_potential_term])

        for kinetic_term in kinetic_terms[potential_term_mode_mask]:
            inner_commutator_term = (
                commutator_ordered_diagonal_coulomb_with_two_body_operator(
                    potential_term, kinetic_term))

            modes_acted_on_by_inner_commutator = set()
            for inner_commutator_action in inner_commutator_term.terms:
                modes_acted_on_by_inner_commutator.update(
                    set(operator[0] for operator in inner_commutator_action))

            if not modes_acted_on_by_inner_commutator:
                continue

            inner_commutator_mode_mask = numpy.logical_or.reduce([
                potential_mask[mode]
                for mode in modes_acted_on_by_inner_commutator
            ])

            # halved_potential_terms for T+V order, potential_terms for V+T
            for outer_potential_term in outer_potential_terms[
                    inner_commutator_mode_mask]:
                commutator_ordered_diagonal_coulomb_with_two_body_operator(
                    outer_potential_term,
                    inner_commutator_term,
                    prior_terms=error_operator)

            inner_commutator_mode_mask = numpy.logical_or.reduce([
                kinetic_mask[qubit]
                for qubit in modes_acted_on_by_inner_commutator
            ])

            # kinetic_terms for T+V order, halved_kinetic_terms for V+T
            for outer_kinetic_term in outer_kinetic_terms[
                    inner_commutator_mode_mask]:
                commutator_ordered_diagonal_coulomb_with_two_body_operator(
                    outer_kinetic_term,
                    inner_commutator_term,
                    prior_terms=error_operator)

    # Divide by 12 to match the error operator definition.
    # If order='V+T', also flip the sign to account for inner_commutator_term
    # not flipping between the different orderings.
    if order == 'T+V':
        error_operator /= 12.0
    else:
        error_operator /= -12.0

    return error_operator


def fermionic_swap_trotter_error_operator_diagonal_two_body(
        hamiltonian, external_potential_at_end=False):
    """Compute the fermionic swap network Trotter error of a diagonal
    two-body Hamiltonian.

    Args:
        hamiltonian (FermionOperator): The diagonal Coulomb Hamiltonian to
                                       compute the Trotter error for.

    Returns:
        error_operator: The second-order Trotter error operator.

    Notes:
        Follows Eq 9 of Poulin et al., arXiv:1406.4920, applied to the
        Trotter step detailed in Kivlichan et al., arxiv:1711.04789.
    """
    single_terms = numpy.array(
        simulation_ordered_grouped_low_depth_terms_with_info(
            hamiltonian,
            external_potential_at_end=external_potential_at_end)[0])

    # Cache the halved terms for use in the second commutator.
    halved_single_terms = single_terms / 2.0

    term_mode_mask = bit_mask_of_modes_acted_on_by_fermionic_terms(
        single_terms, count_qubits(hamiltonian))

    error_operator = FermionOperator.zero()

    for beta, term_beta in enumerate(single_terms):
        modes_acted_on_by_term_beta = set()
        for beta_action in term_beta.terms:
            modes_acted_on_by_term_beta.update(
                set(operator[0] for operator in beta_action))

        beta_mode_mask = numpy.logical_or.reduce(
            [term_mode_mask[mode] for mode in modes_acted_on_by_term_beta])

        # alpha_prime indices that could have a nonzero commutator, i.e.
        # there's overlap between the modes the corresponding terms act on.
        valid_alpha_primes = numpy.where(beta_mode_mask)[0]

        # Only alpha_prime < beta enters the error operator; filter for this.
        valid_alpha_primes = valid_alpha_primes[valid_alpha_primes < beta]

        for alpha_prime in valid_alpha_primes:
            term_alpha_prime = single_terms[alpha_prime]

            inner_commutator_term = (
                commutator_ordered_diagonal_coulomb_with_two_body_operator(
                    term_beta, term_alpha_prime))

            modes_acted_on_by_inner_commutator = set()
            for inner_commutator_action in inner_commutator_term.terms:
                modes_acted_on_by_inner_commutator.update(
                    set(operator[0] for operator in inner_commutator_action))

            # If the inner commutator has no action, the commutator is zero.
            if not modes_acted_on_by_inner_commutator:
                continue

            inner_commutator_mask = numpy.logical_or.reduce([
                term_mode_mask[mode]
                for mode in modes_acted_on_by_inner_commutator
            ])

            # alpha indices that could have a nonzero commutator.
            valid_alphas = numpy.where(inner_commutator_mask)[0]
            # Filter so alpha <= beta in the double commutator.
            valid_alphas = valid_alphas[valid_alphas <= beta]

            for alpha in valid_alphas:
                # If alpha = beta, only use half the term.
                if alpha != beta:
                    outer_term_alpha = single_terms[alpha]
                else:
                    outer_term_alpha = halved_single_terms[alpha]

                # Add the partial double commutator to the error operator.
                commutator_ordered_diagonal_coulomb_with_two_body_operator(
                    outer_term_alpha,
                    inner_commutator_term,
                    prior_terms=error_operator)

    # Divide by 12 to match the error operator definition.
    error_operator /= 12.0
    return error_operator



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import itertools
from typing import cast, Tuple

import numpy as np
import pytest
import scipy.linalg as la
import sympy
import cirq
import cirq.contrib.acquaintance as cca

import openfermion
from openfermion.circuits.gates.fermionic_simulation import (
    sum_of_interaction_operator_gate_generators,
    state_swap_eigen_component,
)


def test_state_swap_eigen_component_args():
    with pytest.raises(TypeError):
        state_swap_eigen_component(0, '12', 1)
    with pytest.raises(ValueError):
        state_swap_eigen_component('01', '01', 1)
    with pytest.raises(ValueError):
        state_swap_eigen_component('01', '10', 0)
    with pytest.raises(ValueError):
        state_swap_eigen_component('01', '100', 1)
    with pytest.raises(ValueError):
        state_swap_eigen_component('01', 'ab', 1)


@pytest.mark.parametrize('index_pair,n_qubits', [
    ((0, 1), 2),
    ((0, 3), 2),
])
def test_state_swap_eigen_component(index_pair, n_qubits):
    state_pair = tuple(format(i, '0' + str(n_qubits) + 'b') for i in index_pair)
    i, j = index_pair
    dim = 2**n_qubits
    for sign in (-1, 1):
        actual_component = state_swap_eigen_component(state_pair[0],
                                                      state_pair[1], sign)
        expected_component = np.zeros((dim, dim))
        expected_component[i, i] = expected_component[j, j] = 0.5
        expected_component[i, j] = expected_component[j, i] = sign * 0.5
        assert np.allclose(actual_component, expected_component)


@pytest.mark.parametrize('n_modes, seed',
                         [(7, np.random.randint(1 << 30)) for _ in range(2)])
def test_interaction_operator_interconversion(n_modes, seed):
    operator = openfermion.random_interaction_operator(n_modes,
                                                       real=False,
                                                       seed=seed)
    gates = openfermion.fermionic_simulation_gates_from_interaction_operator(
        operator)
    other_operator = sum_of_interaction_operator_gate_generators(n_modes, gates)
    operator = openfermion.normal_ordered(operator)
    other_operator = openfermion.normal_ordered(other_operator)
    assert operator == other_operator


def test_interaction_operator_from_bad_gates():
    for gates in [{(): 'bad'}, {(0,): cirq.X}]:
        with pytest.raises(TypeError):
            sum_of_interaction_operator_gate_generators(5, gates)


def random_real(size=None, mag=20):
    return np.random.uniform(-mag, mag, size)


def random_complex(size=None, mag=20):
    return random_real(size, mag) + 1j * random_real(size, mag)


def random_fermionic_simulation_gate(order):
    exponent = random_real()
    if order == 2:
        weights = (random_complex(), random_real())
        return openfermion.QuadraticFermionicSimulationGate(weights,
                                                            exponent=exponent)
    weights = random_complex(3)
    if order == 3:
        return openfermion.CubicFermionicSimulationGate(weights,
                                                        exponent=exponent)
    if order == 4:
        return openfermion.QuarticFermionicSimulationGate(weights,
                                                          exponent=exponent)


def assert_symbolic_decomposition_consistent(gate):
    expected_unitary = cirq.unitary(gate)

    weights = tuple(sympy.Symbol(f'w{i}') for i in range(gate.num_weights()))
    exponent = sympy.Symbol('t')
    symbolic_gate = type(gate)(weights, exponent=exponent)
    qubits = cirq.LineQubit.range(gate.num_qubits())
    circuit = cirq.Circuit(symbolic_gate._decompose_(qubits))
    resolver = {'t': gate.exponent}
    for i, w in enumerate(gate.weights):
        resolver[f'w{i}'] = w
    resolved_circuit = cirq.resolve_parameters(circuit, resolver)
    decomp_unitary = resolved_circuit.unitary(qubit_order=qubits)

    assert np.allclose(expected_unitary, decomp_unitary)


def assert_generators_consistent(gate):
    qubit_generator = gate.qubit_generator_matrix
    qubit_generator_from_fermion_generator = (super(
        type(gate), gate).qubit_generator_matrix)
    assert np.allclose(qubit_generator, qubit_generator_from_fermion_generator)


def assert_resolution_consistent(gate):
    weight_names = [f'w{i}' for i in range(gate.num_weights())]
    weight_params = [sympy.Symbol(w) for w in weight_names]
    resolver = dict(zip(weight_names, gate.weights))
    resolver['s'] = gate._global_shift
    resolver['e'] = gate._exponent
    symbolic_gate = type(gate)(weight_params,
                               exponent=sympy.Symbol('e'),
                               global_shift=sympy.Symbol('s'))
    resolved_gate = cirq.resolve_parameters(symbolic_gate, resolver)
    assert resolved_gate == gate


def assert_fswap_consistent(gate):
    gate = gate.__copy__()
    n_qubits = gate.num_qubits()
    for i in range(n_qubits - 1):
        fswap = cirq.kron(np.eye(1 << i), cirq.unitary(openfermion.FSWAP),
                          np.eye(1 << (n_qubits - i - 2)))
        assert fswap.shape == (1 << n_qubits,) * 2
        generator = gate.qubit_generator_matrix
        fswapped_generator = np.linalg.multi_dot([fswap, generator, fswap])
        gate.fswap(i)
        assert np.allclose(gate.qubit_generator_matrix, fswapped_generator)
    for i in (-1, n_qubits):
        with pytest.raises(ValueError):
            gate.fswap(i)


def assert_permute_consistent(gate):
    gate = gate.__copy__()
    n_qubits = gate.num_qubits()
    qubits = cirq.LineQubit.range(n_qubits)
    for pos in itertools.permutations(range(n_qubits)):
        permuted_gate = gate.__copy__()
        gate.permute(pos)
        assert permuted_gate.permuted(pos) == gate
        actual_unitary = cirq.unitary(permuted_gate)

        ops = [
            cca.LinearPermutationGate(n_qubits, dict(zip(range(n_qubits), pos)),
                                      openfermion.FSWAP)(*qubits),
            gate(*qubits),
            cca.LinearPermutationGate(n_qubits, dict(zip(pos, range(n_qubits))),
                                      openfermion.FSWAP)(*qubits)
        ]
        circuit = cirq.Circuit(ops)
        expected_unitary = cirq.unitary(circuit)
        assert np.allclose(actual_unitary, expected_unitary)

    with pytest.raises(ValueError):
        gate.permute(range(1, n_qubits))
    with pytest.raises(ValueError):
        gate.permute([1] * n_qubits)


def assert_interaction_operator_consistent(gate):
    interaction_op = gate.interaction_operator_generator()
    other_gate = gate.from_interaction_operator(operator=interaction_op)
    if other_gate is None:
        assert np.allclose(gate.weights, 0)
    else:
        assert cirq.approx_eq(gate, other_gate)
    interaction_op = openfermion.normal_ordered(interaction_op)
    other_interaction_op = openfermion.InteractionOperator.zero(
        interaction_op.n_qubits)
    super(type(gate),
          gate).interaction_operator_generator(operator=other_interaction_op)
    other_interaction_op = openfermion.normal_ordered(interaction_op)
    assert interaction_op == other_interaction_op

    other_interaction_op = super(type(gate),
                                 gate).interaction_operator_generator()
    other_interaction_op = openfermion.normal_ordered(interaction_op)
    assert interaction_op == other_interaction_op


random_quadratic_gates = [random_fermionic_simulation_gate(2) for _ in range(5)]
manual_quadratic_gates = [
    openfermion.QuadraticFermionicSimulationGate(weights)
    for weights in [cast(Tuple[float, float], (1, 1)), (1, 0), (0, 1), (0, 0)]
]
quadratic_gates = random_quadratic_gates + manual_quadratic_gates
cubic_gates = ([openfermion.CubicFermionicSimulationGate()] +
               [random_fermionic_simulation_gate(3) for _ in range(5)])
quartic_gates = ([openfermion.QuarticFermionicSimulationGate()] +
                 [random_fermionic_simulation_gate(4) for _ in range(5)])
gates = quadratic_gates + cubic_gates + quartic_gates


@pytest.mark.parametrize('gate', gates)
def test_fermionic_simulation_gate(gate):
    openfermion.testing.assert_implements_consistent_protocols(gate)

    generator = gate.qubit_generator_matrix
    expected_unitary = la.expm(-1j * gate.exponent * generator)
    actual_unitary = cirq.unitary(gate)
    assert np.allclose(expected_unitary, actual_unitary)

    assert_fswap_consistent(gate)
    assert_permute_consistent(gate)
    assert_generators_consistent(gate)
    assert_resolution_consistent(gate)
    assert_interaction_operator_consistent(gate)

    assert gate.num_weights() == super(type(gate), gate).num_weights()


@pytest.mark.parametrize('weights', list(np.random.rand(10, 3)) + [(1, 0, 1)])
def test_weights_and_exponent(weights):
    exponents = np.linspace(-1, 1, 8)
    gates = tuple(
        openfermion.QuarticFermionicSimulationGate(
            weights / exponent, exponent=exponent, absorb_exponent=True)
        for exponent in exponents)

    for g1, g2 in itertools.combinations(gates, 2):
        assert cirq.approx_eq(g1, g2, atol=1e-100)

    for i, (gate, exponent) in enumerate(zip(gates, exponents)):
        assert gate.exponent == 1
        new_exponent = exponents[-i]
        new_gate = gate._with_exponent(new_exponent)
        assert new_gate.exponent == new_exponent


def test_zero_weights():
    for gate_type in [
            openfermion.QuadraticFermionicSimulationGate,
            openfermion.CubicFermionicSimulationGate,
            openfermion.QuarticFermionicSimulationGate
    ]:
        weights = (0,) * gate_type.num_weights()
        gate = gate_type(weights)
        n_qubits = gate.num_qubits()

        assert np.allclose(cirq.unitary(gate), np.eye(2**n_qubits))
        cirq.testing.assert_decompose_is_consistent_with_unitary(gate)

        operator = openfermion.InteractionOperator.zero(n_qubits)
        assert gate_type.from_interaction_operator(operator=operator) is None


@pytest.mark.parametrize(
    'weights,exponent',
    [((np.random.uniform(-5, 5) + 1j * np.random.uniform(-5, 5),
       np.random.uniform(-5, 5)), np.random.uniform(-5, 5)) for _ in range(5)])
def test_quadratic_fermionic_simulation_gate_unitary(weights, exponent):
    generator = np.zeros((4, 4), dtype=np.complex128)
    # w0 |10><01| + h.c.
    generator[2, 1] = weights[0]
    generator[1, 2] = weights[0].conjugate()
    # w1 |11><11|
    generator[3, 3] = weights[1]
    expected_unitary = la.expm(-1j * exponent * generator)

    gate = openfermion.QuadraticFermionicSimulationGate(weights,
                                                        exponent=exponent)
    actual_unitary = cirq.unitary(gate)

    assert np.allclose(expected_unitary, actual_unitary)

    symbolic_gate = (openfermion.QuadraticFermionicSimulationGate(
        (sympy.Symbol('w0'), sympy.Symbol('w1')), exponent=sympy.Symbol('t')))
    qubits = cirq.LineQubit.range(2)
    circuit = cirq.Circuit(symbolic_gate._decompose_(qubits))
    resolver = {'w0': weights[0], 'w1': weights[1], 't': exponent}
    resolved_circuit = cirq.resolve_parameters(circuit, resolver)
    decomp_unitary = resolved_circuit.unitary(qubit_order=qubits)

    assert np.allclose(expected_unitary, decomp_unitary)


@pytest.mark.parametrize('gate', random_quadratic_gates)
def test_quadratic_fermionic_simulation_gate_symbolic_decompose(gate):
    assert_symbolic_decomposition_consistent(gate)


def test_cubic_fermionic_simulation_gate_equality():
    eq = cirq.testing.EqualsTester()
    eq.add_equality_group(
        openfermion.CubicFermionicSimulationGate()**0.5,
        openfermion.CubicFermionicSimulationGate((1,) * 3, exponent=0.5),
        openfermion.CubicFermionicSimulationGate((0.5,) * 3),
    )
    eq.add_equality_group(openfermion.CubicFermionicSimulationGate((1j, 0, 0)),)
    eq.add_equality_group(
        openfermion.CubicFermionicSimulationGate((sympy.Symbol('s'), 0, 0),
                                                 exponent=2),
        openfermion.CubicFermionicSimulationGate((2 * sympy.Symbol('s'), 0, 0),
                                                 exponent=1),
    )
    eq.add_equality_group(
        openfermion.CubicFermionicSimulationGate((0, 0.7, 0), global_shift=2),
        openfermion.CubicFermionicSimulationGate((0, 0.35, 0),
                                                 global_shift=1,
                                                 exponent=2),
    )
    eq.add_equality_group(
        openfermion.CubicFermionicSimulationGate((1, 1, 1)),
        openfermion.CubicFermionicSimulationGate(((1 + 2 * np.pi), 1, 1)),
    )


@pytest.mark.parametrize('exponent,control',
                         itertools.product([0, 1, -1, 0.25, -0.5, 0.1],
                                           [0, 1, 2]))
def test_cubic_fermionic_simulation_gate_consistency_special(exponent, control):
    weights = tuple(np.eye(1, 3, control)[0] * 0.5 * np.pi)
    general_gate = openfermion.CubicFermionicSimulationGate(weights,
                                                            exponent=exponent)
    general_unitary = cirq.unitary(general_gate)

    indices = np.dot(list(itertools.product((0, 1), repeat=3)),
                     (2**np.roll(np.arange(3), -control))[::-1])
    special_gate = cirq.ControlledGate(cirq.ISWAP**-exponent)
    special_unitary = (
        cirq.unitary(special_gate)[indices[:, np.newaxis], indices])

    assert np.allclose(general_unitary, special_unitary)


@pytest.mark.parametrize(
    'weights,exponent',
    [(np.random.uniform(-5, 5, 3) + 1j * np.random.uniform(-5, 5, 3),
      np.random.uniform(-5, 5)) for _ in range(5)])
def test_cubic_fermionic_simulation_gate_consistency_docstring(
        weights, exponent):
    generator = np.zeros((8, 8), dtype=np.complex128)
    # w0 |110><101| + h.c.
    generator[6, 5] = weights[0]
    generator[5, 6] = weights[0].conjugate()
    # w1 |110><011| + h.c.
    generator[6, 3] = weights[1]
    generator[3, 6] = weights[1].conjugate()
    # w2 |101><011| + h.c.
    generator[5, 3] = weights[2]
    generator[3, 5] = weights[2].conjugate()
    expected_unitary = la.expm(-1j * exponent * generator)

    gate = openfermion.CubicFermionicSimulationGate(weights, exponent=exponent)
    actual_unitary = cirq.unitary(gate)

    assert np.allclose(expected_unitary, actual_unitary)


def test_quartic_fermionic_simulation_consistency():
    openfermion.testing.assert_implements_consistent_protocols(
        openfermion.QuarticFermionicSimulationGate())


quartic_fermionic_simulation_simulator_test_cases = [
    (openfermion.QuarticFermionicSimulationGate(
        (0, 0, 0)), 1., np.ones(16) / 4., np.ones(16) / 4., 5e-6),
    (openfermion.QuarticFermionicSimulationGate((0.2, -0.1, 0.7)), 0.,
     np.array([1, -1, -1, -1, -1, -1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1]) / 4.,
     np.array([1, -1, -1, -1, -1, -1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1]) / 4.,
     5e-6),
    (openfermion.QuarticFermionicSimulationGate((0.2, -0.1, 0.7)), 0.3,
     np.array([1, -1, -1, -1, -1, -1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1]) / 4.,
     np.array([
         1, -1, -1, -np.exp(0.21j), -1, -np.exp(-0.03j),
         np.exp(-0.06j), 1, 1,
         np.exp(-0.06j),
         np.exp(-0.03j), 1,
         np.exp(0.21j), 1, 1, 1
     ]) / 4., 5e-6),
    (openfermion.QuarticFermionicSimulationGate((1. / 3, 0, 0)), 1.,
     np.array([0, 0, 0, 0, 0, 0, 1., 0, 0, 1., 0, 0, 0, 0, 0, 0]) / np.sqrt(2),
     np.array([0, 0, 0, 0, 0, 0, 1., 0, 0, 1., 0, 0, 0, 0, 0, 0]) / np.sqrt(2),
     5e-6),
    (openfermion.QuarticFermionicSimulationGate((0, np.pi / 3, 0)), 1.,
     np.array([1., 1., 0, 0, 0, 1., 0, 0, 0, 0., -1., 0, 0, 0, 0, 0]) / 2.,
     np.array([
         1., 1., 0, 0, 0, -np.exp(4j * np.pi / 3), 0, 0, 0, 0.,
         -np.exp(1j * np.pi / 3), 0, 0, 0, 0, 0
     ]) / 2., 5e-6),
    (openfermion.QuarticFermionicSimulationGate((0, 0, -np.pi / 2)), 1.,
     np.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1., 0, 0,
               0]), np.array([0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                              0]), 5e-6),
    (openfermion.QuarticFermionicSimulationGate((0, 0, -0.25 * np.pi)), 1.,
     np.array([0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]),
     np.array([0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1j, 0, 0, 0]) / np.sqrt(2),
     5e-6),
    (openfermion.QuarticFermionicSimulationGate(
        (-np.pi / 4, np.pi / 6, -np.pi / 2)), 1.,
     np.array([0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 1, 0, 0, 0]) / np.sqrt(3),
     np.array([
         0, 0, 0, 1j, 0, -1j / 2., 1 / np.sqrt(2), 0, 0, 1j / np.sqrt(2),
         np.sqrt(3) / 2, 0, 0, 0, 0, 0
     ]) / np.sqrt(3), 5e-6),
]


@pytest.mark.parametrize('gate, exponent, initial_state, correct_state, atol',
                         quartic_fermionic_simulation_simulator_test_cases)
def test_quartic_fermionic_simulation_on_simulator(gate, exponent,
                                                   initial_state, correct_state,
                                                   atol):

    a, b, c, d = cirq.LineQubit.range(4)
    circuit = cirq.Circuit(gate(a, b, c, d)**exponent)
    result = circuit.final_state_vector(initial_state=initial_state)
    cirq.testing.assert_allclose_up_to_global_phase(result,
                                                    correct_state,
                                                    atol=atol)


def test_quartic_fermionic_simulation_eq():
    eq = cirq.testing.EqualsTester()

    eq.add_equality_group(
        openfermion.QuarticFermionicSimulationGate((1.2, 0.4, -0.4),
                                                   exponent=0.5),
        openfermion.QuarticFermionicSimulationGate((0.3, 0.1, -0.1),
                                                   exponent=2),
        openfermion.QuarticFermionicSimulationGate((-0.6, -0.2, 0.2),
                                                   exponent=-1),
        openfermion.QuarticFermionicSimulationGate((0.6, 0.2, 2 * np.pi - 0.2)),
    )

    eq.add_equality_group(
        openfermion.QuarticFermionicSimulationGate((-0.6, 0.0, 0.3),
                                                   exponent=0.5))

    eq.make_equality_group(lambda: openfermion.QuarticFermionicSimulationGate(
        (0.1, -0.3, 0.0), exponent=0.0))
    eq.make_equality_group(lambda: openfermion.QuarticFermionicSimulationGate(
        (1., -1., 0.5), exponent=0.75))


def test_quadratic_fermionic_simulation_gate_text_diagram():
    gate = openfermion.QuadraticFermionicSimulationGate((1, 1))
    a, b, c = cirq.LineQubit.range(3)
    circuit = cirq.Circuit([gate(a, b), gate(b, c)])

    assert super(type(gate), gate).wire_symbol(False) == type(gate).__name__
    assert (super(type(gate), gate)._diagram_exponent(
        cirq.CircuitDiagramInfoArgs.UNINFORMED_DEFAULT) == gate._exponent)

    expected_text_diagram = """
0: ───↓↑(1, 1)──────────────
      │
1: ───↓↑─────────↓↑(1, 1)───
                 │
2: ──────────────↓↑─────────
""".strip()
    cirq.testing.assert_has_diagram(circuit, expected_text_diagram)

    expected_text_diagram = """
0: ---a*a(1, 1)---------------
      |
1: ---a*a---------a*a(1, 1)---
                  |
2: ---------------a*a---------
""".strip()
    cirq.testing.assert_has_diagram(circuit,
                                    expected_text_diagram,
                                    use_unicode_characters=False)


def test_cubic_fermionic_simulation_gate_text_diagram():
    gate = openfermion.CubicFermionicSimulationGate((1, 1, 1))
    qubits = cirq.LineQubit.range(5)
    circuit = cirq.Circuit([gate(*qubits[:3]), gate(*qubits[2:5])])

    assert super(type(gate), gate).wire_symbol(False) == type(gate).__name__
    assert (super(type(gate), gate)._diagram_exponent(
        cirq.CircuitDiagramInfoArgs.UNINFORMED_DEFAULT) == gate._exponent)

    expected_text_diagram = """
0: ───↕↓↑(1, 1, 1)──────────────────
      │
1: ───↕↓↑───────────────────────────
      │
2: ───↕↓↑────────────↕↓↑(1, 1, 1)───
                     │
3: ──────────────────↕↓↑────────────
                     │
4: ──────────────────↕↓↑────────────
""".strip()
    cirq.testing.assert_has_diagram(circuit, expected_text_diagram)

    expected_text_diagram = """
0: ---na*a(1, 1, 1)-------------------
      |
1: ---na*a----------------------------
      |
2: ---na*a------------na*a(1, 1, 1)---
                      |
3: -------------------na*a------------
                      |
4: -------------------na*a------------
""".strip()
    cirq.testing.assert_has_diagram(circuit,
                                    expected_text_diagram,
                                    use_unicode_characters=False)


test_weights = [1.0, 0.5, 0.25, 0.1, 0.0, -0.5]


@pytest.mark.parametrize('weights',
                         itertools.chain(
                             itertools.product(test_weights, repeat=3),
                             np.random.rand(10, 3)))
def test_quartic_fermionic_simulation_decompose(weights):
    cirq.testing.assert_decompose_is_consistent_with_unitary(
        openfermion.QuarticFermionicSimulationGate(weights))


@pytest.mark.parametrize(
    'weights,exponent',
    [(np.random.uniform(-5, 5, 3) + 1j * np.random.uniform(-5, 5, 3),
      np.random.uniform(-5, 5)) for _ in range(5)])
def test_quartic_fermionic_simulation_unitary(weights, exponent):
    generator = np.zeros((1 << 4,) * 2, dtype=np.complex128)

    # w0 |1001><0110| + h.c.
    generator[9, 6] = weights[0]
    generator[6, 9] = weights[0].conjugate()
    # w1 |1010><0101| + h.c.
    generator[10, 5] = weights[1]
    generator[5, 10] = weights[1].conjugate()
    # w2 |1100><0011| + h.c.
    generator[12, 3] = weights[2]
    generator[3, 12] = weights[2].conjugate()
    expected_unitary = la.expm(-1j * exponent * generator)

    gate = openfermion.QuarticFermionicSimulationGate(weights,
                                                      exponent=exponent)
    actual_unitary = cirq.unitary(gate)

    assert np.allclose(expected_unitary, actual_unitary)


def test_quartic_fermionic_simulation_gate_text_diagram():
    gate = openfermion.QuarticFermionicSimulationGate((1, 1, 1))
    qubits = cirq.LineQubit.range(6)
    circuit = cirq.Circuit([gate(*qubits[:4]), gate(*qubits[-4:])])

    assert super(type(gate), gate).wire_symbol(False) == type(gate).__name__
    for G in (gate, gate._with_exponent('e')):
        assert (super(type(G), G)._diagram_exponent(
            cirq.CircuitDiagramInfoArgs.UNINFORMED_DEFAULT) == G._exponent)

    expected_text_diagram = """
0: ───⇊⇈(1, 1, 1)─────────────────
      │
1: ───⇊⇈──────────────────────────
      │
2: ───⇊⇈────────────⇊⇈(1, 1, 1)───
      │             │
3: ───⇊⇈────────────⇊⇈────────────
                    │
4: ─────────────────⇊⇈────────────
                    │
5: ─────────────────⇊⇈────────────
""".strip()
    cirq.testing.assert_has_diagram(circuit, expected_text_diagram)

    expected_text_diagram = """
0: ---a*a*aa(1, 1, 1)---------------------
      |
1: ---a*a*aa------------------------------
      |
2: ---a*a*aa------------a*a*aa(1, 1, 1)---
      |                 |
3: ---a*a*aa------------a*a*aa------------
                        |
4: ---------------------a*a*aa------------
                        |
5: ---------------------a*a*aa------------
""".strip()
    cirq.testing.assert_has_diagram(circuit,
                                    expected_text_diagram,
                                    use_unicode_characters=False)


@pytest.mark.parametrize(
    'weights,exponent',
    [(np.random.uniform(-5, 5, 3) + 1j * np.random.uniform(-5, 5, 3),
      np.random.uniform(-5, 5)) for _ in range(5)])
def test_quartic_fermionic_simulation_apply_unitary(weights, exponent):
    gate = openfermion.QuarticFermionicSimulationGate(weights,
                                                      exponent=exponent)
    cirq.testing.assert_has_consistent_apply_unitary(gate, atol=5e-6)



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Tests for commutators.py."""
import unittest

import numpy

from openfermion.ops.operators import (FermionOperator, QubitOperator,
                                       BosonOperator, QuadOperator)
from openfermion.transforms import jordan_wigner
from openfermion.utils import hermitian_conjugated
from openfermion.transforms.opconversions import normal_ordered
from openfermion.linalg.sparse_tools import pauli_matrix_map
from openfermion.utils.commutators import (
    commutator, anticommutator, double_commutator,
    trivially_commutes_dual_basis, trivially_double_commutes_dual_basis,
    trivially_double_commutes_dual_basis_using_term_info)


class CommutatorTest(unittest.TestCase):

    def setUp(self):
        self.fermion_term = FermionOperator('1^ 2^ 3 4', -3.17)
        self.fermion_operator = self.fermion_term + hermitian_conjugated(
            self.fermion_term)
        self.boson_term = BosonOperator('1^ 2^ 3 4', -3.17)
        self.boson_operator = self.boson_term + hermitian_conjugated(
            self.boson_term)
        self.quad_term = QuadOperator('q0 p0 q1 p0 p0', -3.17)
        self.quad_operator = self.quad_term + hermitian_conjugated(
            self.quad_term)
        self.qubit_operator = jordan_wigner(self.fermion_operator)

    def test_commutes_identity(self):
        com = commutator(FermionOperator.identity(),
                         FermionOperator('2^ 3', 2.3))
        self.assertEqual(com, FermionOperator.zero())

        com = commutator(BosonOperator.identity(), BosonOperator('2^ 3', 2.3))
        self.assertTrue(com == BosonOperator.zero())

        com = commutator(QuadOperator.identity(), QuadOperator('q2 p3', 2.3))
        self.assertTrue(com == QuadOperator.zero())

    def test_commutes_no_intersection(self):
        com = commutator(FermionOperator('2^ 3'), FermionOperator('4^ 5^ 3'))
        com = normal_ordered(com)
        self.assertEqual(com, FermionOperator.zero())

        com = commutator(BosonOperator('2^ 3'), BosonOperator('4^ 5^ 3'))
        com = normal_ordered(com)
        self.assertTrue(com == BosonOperator.zero())

        com = commutator(QuadOperator('q2 p3'), QuadOperator('q4 q5 p3'))
        com = normal_ordered(com)
        self.assertTrue(com == QuadOperator.zero())

    def test_commutes_number_operators(self):
        com = commutator(FermionOperator('4^ 3^ 4 3'), FermionOperator('2^ 2'))
        com = normal_ordered(com)
        self.assertEqual(com, FermionOperator.zero())

        com = commutator(BosonOperator('4^ 3^ 4 3'), BosonOperator('2^ 2'))
        com = normal_ordered(com)
        self.assertTrue(com == BosonOperator.zero())

    def test_commutator_hopping_operators(self):
        com = commutator(3 * FermionOperator('1^ 2'), FermionOperator('2^ 3'))
        com = normal_ordered(com)
        self.assertEqual(com, FermionOperator('1^ 3', 3))

        com = commutator(3 * BosonOperator('1^ 2'), BosonOperator('2^ 3'))
        com = normal_ordered(com)
        self.assertTrue(com == BosonOperator('1^ 3', 3))

    def test_commutator_hopping_with_single_number(self):
        com = commutator(FermionOperator('1^ 2', 1j), FermionOperator('1^ 1'))
        com = normal_ordered(com)
        self.assertEqual(com, -FermionOperator('1^ 2') * 1j)

    def test_commutator_hopping_with_double_number_one_intersection(self):
        com = commutator(FermionOperator('1^ 3'), FermionOperator('3^ 2^ 3 2'))
        com = normal_ordered(com)
        self.assertEqual(com, -FermionOperator('2^ 1^ 3 2'))

    def test_commutator_hopping_with_double_number_two_intersections(self):
        com = commutator(FermionOperator('2^ 3'), FermionOperator('3^ 2^ 3 2'))
        com = normal_ordered(com)
        self.assertEqual(com, FermionOperator.zero())

    def test_commutator(self):
        operator_a = FermionOperator('')
        self.assertEqual(FermionOperator.zero(),
                         commutator(operator_a, self.fermion_operator))
        operator_b = QubitOperator('X1 Y2')
        self.assertEqual(commutator(self.qubit_operator, operator_b),
                         (self.qubit_operator * operator_b -
                          operator_b * self.qubit_operator))

    def test_canonical_boson_commutation_relations(self):
        op_1 = BosonOperator('3')
        op_1_dag = BosonOperator('3^')
        op_2 = BosonOperator('4')
        op_2_dag = BosonOperator('4^')
        zero = BosonOperator()
        one = BosonOperator('')

        self.assertTrue(one == normal_ordered(commutator(op_1, op_1_dag)))
        self.assertTrue(zero == normal_ordered(commutator(op_1, op_2)))
        self.assertTrue(zero == normal_ordered(commutator(op_1, op_2_dag)))
        self.assertTrue(zero == normal_ordered(commutator(op_1_dag, op_2)))
        self.assertTrue(zero == normal_ordered(commutator(op_1_dag, op_2_dag)))
        self.assertTrue(one == normal_ordered(commutator(op_2, op_2_dag)))

    def test_canonical_quad_commutation_relations(self):
        q1 = QuadOperator('q3')
        p1 = QuadOperator('p3')
        q2 = QuadOperator('q4')
        p2 = QuadOperator('p4')
        zero = QuadOperator()
        one = QuadOperator('')
        hbar = 2.

        self.assertTrue(1j * hbar *
                        one == normal_ordered(commutator(q1, p1), hbar))
        self.assertTrue(zero == normal_ordered(commutator(q1, q2), hbar))
        self.assertTrue(zero == normal_ordered(commutator(q1, p2), hbar))
        self.assertTrue(zero == normal_ordered(commutator(p1, q2), hbar))
        self.assertTrue(zero == normal_ordered(commutator(p1, p2), hbar))
        self.assertTrue(1j * hbar *
                        one == normal_ordered(commutator(q2, p2), hbar))

    def test_ndarray_input(self):
        """Test when the inputs are numpy arrays."""
        X = pauli_matrix_map['X'].toarray()
        Y = pauli_matrix_map['Y'].toarray()
        Z = pauli_matrix_map['Z'].toarray()
        self.assertTrue(numpy.allclose(commutator(X, Y), 2.j * Z))

    def test_commutator_operator_a_bad_type(self):
        with self.assertRaises(TypeError):
            commutator(1, self.fermion_operator)

    def test_commutator_operator_b_bad_type(self):
        with self.assertRaises(TypeError):
            commutator(self.qubit_operator, "hello")

    def test_commutator_not_same_type(self):
        with self.assertRaises(TypeError):
            commutator(self.fermion_operator, self.qubit_operator)


class AnticommutatorTest(unittest.TestCase):

    def test_canonical_anticommutation_relations(self):
        op_1 = FermionOperator('3')
        op_1_dag = FermionOperator('3^')
        op_2 = FermionOperator('4')
        op_2_dag = FermionOperator('4^')
        zero = FermionOperator.zero()
        one = FermionOperator.identity()

        self.assertEqual(one, normal_ordered(anticommutator(op_1, op_1_dag)))
        self.assertEqual(zero, normal_ordered(anticommutator(op_1, op_2)))
        self.assertEqual(zero, normal_ordered(anticommutator(op_1, op_2_dag)))
        self.assertEqual(zero, normal_ordered(anticommutator(op_1_dag, op_2)))
        self.assertEqual(zero,
                         normal_ordered(anticommutator(op_1_dag, op_2_dag)))
        self.assertEqual(one, normal_ordered(anticommutator(op_2, op_2_dag)))

    def test_ndarray_input(self):
        """Test when the inputs are numpy arrays."""
        X = pauli_matrix_map['X'].toarray()
        Y = pauli_matrix_map['Y'].toarray()
        zero = numpy.zeros((2, 2))
        self.assertTrue(numpy.allclose(anticommutator(X, Y), zero))

    def test_anticommutator_not_same_type(self):
        with self.assertRaises(TypeError):
            anticommutator(FermionOperator(), QubitOperator())


class DoubleCommutatorTest(unittest.TestCase):

    def test_double_commutator_no_intersection_with_union_of_second_two(self):
        com = double_commutator(FermionOperator('4^ 3^ 6 5'),
                                FermionOperator('2^ 1 0'),
                                FermionOperator('0^'))
        self.assertEqual(com, FermionOperator.zero())

    def test_double_commutator_more_info_not_hopping(self):
        com = double_commutator(FermionOperator('3^ 2'),
                                FermionOperator('2^ 3') +
                                FermionOperator('3^ 2'),
                                FermionOperator('4^ 2^ 4 2'),
                                indices2=set([2, 3]),
                                indices3=set([2, 4]),
                                is_hopping_operator2=True,
                                is_hopping_operator3=False)
        self.assertEqual(
            com, (FermionOperator('4^ 2^ 4 2') - FermionOperator('4^ 3^ 4 3')))

    def test_double_commtator_more_info_both_hopping(self):
        com = double_commutator(
            FermionOperator('4^ 3^ 4 3'),
            FermionOperator('1^ 2', 2.1) + FermionOperator('2^ 1', 2.1),
            FermionOperator('1^ 3', -1.3) + FermionOperator('3^ 1', -1.3),
            indices2=set([1, 2]),
            indices3=set([1, 3]),
            is_hopping_operator2=True,
            is_hopping_operator3=True)
        self.assertEqual(com, (FermionOperator('4^ 3^ 4 2', 2.73) +
                               FermionOperator('4^ 2^ 4 3', 2.73)))


class TriviallyDoubleCommutesDualBasisUsingTermInfoTest(unittest.TestCase):

    def test_number_operators_trivially_commute(self):
        self.assertTrue(
            trivially_double_commutes_dual_basis_using_term_info(
                indices_alpha=set([1, 2]),
                indices_beta=set([3, 4]),
                indices_alpha_prime=set([2, 3]),
                is_hopping_operator_alpha=False,
                is_hopping_operator_beta=False,
                is_hopping_operator_alpha_prime=False,
                jellium_only=True))

    def test_left_hopping_operator_no_trivial_commutation(self):
        self.assertFalse(
            trivially_double_commutes_dual_basis_using_term_info(
                indices_alpha=set([1, 2]),
                indices_beta=set([3, 4]),
                indices_alpha_prime=set([2, 3]),
                is_hopping_operator_alpha=True,
                is_hopping_operator_beta=True,
                is_hopping_operator_alpha_prime=False,
                jellium_only=True))

    def test_right_hopping_operator_no_trivial_commutation(self):
        self.assertFalse(
            trivially_double_commutes_dual_basis_using_term_info(
                indices_alpha=set([1, 2]),
                indices_beta=set([3, 4]),
                indices_alpha_prime=set([2, 3]),
                is_hopping_operator_alpha=True,
                is_hopping_operator_beta=False,
                is_hopping_operator_alpha_prime=True,
                jellium_only=True))

    def test_alpha_is_hopping_operator_others_number_trivial_commutation(self):
        self.assertTrue(
            trivially_double_commutes_dual_basis_using_term_info(
                indices_alpha=set([1, 2]),
                indices_beta=set([3, 4]),
                indices_alpha_prime=set([2, 3]),
                is_hopping_operator_alpha=True,
                is_hopping_operator_beta=False,
                is_hopping_operator_alpha_prime=False,
                jellium_only=True))

    def test_no_intersection_in_first_commutator_trivially_commutes(self):
        self.assertTrue(
            trivially_double_commutes_dual_basis_using_term_info(
                indices_alpha=set([1, 2]),
                indices_beta=set([3, 4]),
                indices_alpha_prime=set([1, 2]),
                is_hopping_operator_alpha=True,
                is_hopping_operator_beta=True,
                is_hopping_operator_alpha_prime=False,
                jellium_only=True))

    def test_double_intersection_in_first_commutator_trivially_commutes(self):
        self.assertTrue(
            trivially_double_commutes_dual_basis_using_term_info(
                indices_alpha=set([3, 2]),
                indices_beta=set([3, 4]),
                indices_alpha_prime=set([4, 3]),
                is_hopping_operator_alpha=True,
                is_hopping_operator_beta=True,
                is_hopping_operator_alpha_prime=False,
                jellium_only=True))

    def test_single_intersection_in_first_commutator_nontrivial(self):
        self.assertFalse(
            trivially_double_commutes_dual_basis_using_term_info(
                indices_alpha=set([3, 2]),
                indices_beta=set([3, 4]),
                indices_alpha_prime=set([4, 5]),
                is_hopping_operator_alpha=False,
                is_hopping_operator_beta=True,
                is_hopping_operator_alpha_prime=False,
                jellium_only=True))

    def test_no_intersection_between_first_and_other_terms_is_trivial(self):
        self.assertTrue(
            trivially_double_commutes_dual_basis_using_term_info(
                indices_alpha=set([3, 2]),
                indices_beta=set([1, 4]),
                indices_alpha_prime=set([4, 5]),
                is_hopping_operator_alpha=False,
                is_hopping_operator_beta=True,
                is_hopping_operator_alpha_prime=False,
                jellium_only=True))


class TriviallyCommutesDualBasisTest(unittest.TestCase):

    def test_trivially_commutes_no_intersection(self):
        self.assertTrue(
            trivially_commutes_dual_basis(FermionOperator('3^ 2^ 3 2'),
                                          FermionOperator('4^ 1')))

    def test_no_trivial_commute_with_intersection(self):
        self.assertFalse(
            trivially_commutes_dual_basis(FermionOperator('2^ 1'),
                                          FermionOperator('5^ 2^ 5 2')))

    def test_trivially_commutes_both_single_number_operators(self):
        self.assertTrue(
            trivially_commutes_dual_basis(FermionOperator('3^ 3'),
                                          FermionOperator('3^ 3')))

    def test_trivially_commutes_nonintersecting_single_number_operators(self):
        self.assertTrue(
            trivially_commutes_dual_basis(FermionOperator('2^ 2'),
                                          FermionOperator('3^ 3')))

    def test_trivially_commutes_both_double_number_operators(self):
        self.assertTrue(
            trivially_commutes_dual_basis(FermionOperator('3^ 2^ 3 2'),
                                          FermionOperator('3^ 1^ 3 1')))

    def test_trivially_commutes_one_double_number_operators(self):
        self.assertTrue(
            trivially_commutes_dual_basis(FermionOperator('3^ 2^ 3 2'),
                                          FermionOperator('3^ 3')))

    def test_no_trivial_commute_right_hopping_operator(self):
        self.assertFalse(
            trivially_commutes_dual_basis(FermionOperator('3^ 1^ 3 1'),
                                          FermionOperator('3^ 2')))

    def test_no_trivial_commute_left_hopping_operator(self):
        self.assertFalse(
            trivially_commutes_dual_basis(FermionOperator('3^ 2'),
                                          FermionOperator('3^ 3')))

    def test_trivially_commutes_both_hopping_create_same_mode(self):
        self.assertTrue(
            trivially_commutes_dual_basis(FermionOperator('3^ 2'),
                                          FermionOperator('3^ 1')))

    def test_trivially_commutes_both_hopping_annihilate_same_mode(self):
        self.assertTrue(
            trivially_commutes_dual_basis(FermionOperator('4^ 1'),
                                          FermionOperator('3^ 1')))

    def test_trivially_commutes_both_hopping_and_number_on_same_modes(self):
        self.assertTrue(
            trivially_commutes_dual_basis(FermionOperator('4^ 1'),
                                          FermionOperator('4^ 1^ 4 1')))


class TriviallyDoubleCommutesDualBasisTest(unittest.TestCase):

    def test_trivially_double_commutes_no_intersection(self):
        self.assertTrue(
            trivially_double_commutes_dual_basis(FermionOperator('3^ 4'),
                                                 FermionOperator('3^ 2^ 3 2'),
                                                 FermionOperator('4^ 1')))

    def test_no_trivial_double_commute_with_intersection(self):
        self.assertFalse(
            trivially_double_commutes_dual_basis(FermionOperator('4^ 2'),
                                                 FermionOperator('2^ 1'),
                                                 FermionOperator('5^ 2^ 5 2')))

    def test_trivially_double_commutes_both_single_number_operators(self):
        self.assertTrue(
            trivially_double_commutes_dual_basis(FermionOperator('4^ 3'),
                                                 FermionOperator('3^ 3'),
                                                 FermionOperator('3^ 3')))

    def test_trivially_double_commutes_nonintersecting_single_number_ops(self):
        self.assertTrue(
            trivially_double_commutes_dual_basis(FermionOperator('3^ 2'),
                                                 FermionOperator('2^ 2'),
                                                 FermionOperator('3^ 3')))

    def test_trivially_double_commutes_both_double_number_operators(self):
        self.assertTrue(
            trivially_double_commutes_dual_basis(FermionOperator('4^ 3'),
                                                 FermionOperator('3^ 2^ 3 2'),
                                                 FermionOperator('3^ 1^ 3 1')))

    def test_trivially_double_commutes_one_double_number_operators(self):
        self.assertTrue(
            trivially_double_commutes_dual_basis(FermionOperator('4^ 3'),
                                                 FermionOperator('3^ 2^ 3 2'),
                                                 FermionOperator('3^ 3')))

    def test_no_trivial_double_commute_right_hopping_operator(self):
        self.assertFalse(
            trivially_double_commutes_dual_basis(FermionOperator('4^ 3'),
                                                 FermionOperator('3^ 1^ 3 1'),
                                                 FermionOperator('3^ 2')))

    def test_no_trivial_double_commute_left_hopping_operator(self):
        self.assertFalse(
            trivially_double_commutes_dual_basis(FermionOperator('4^ 3'),
                                                 FermionOperator('3^ 2'),
                                                 FermionOperator('3^ 3')))

    def test_trivially_double_commutes_both_hopping_create_same_mode(self):
        self.assertTrue(
            trivially_double_commutes_dual_basis(FermionOperator('3^ 3'),
                                                 FermionOperator('3^ 2'),
                                                 FermionOperator('3^ 1')))

    def test_trivially_double_commutes_both_hopping_annihilate_same_mode(self):
        self.assertTrue(
            trivially_double_commutes_dual_basis(FermionOperator('1^ 1'),
                                                 FermionOperator('4^ 1'),
                                                 FermionOperator('3^ 1')))

    def test_trivially_double_commutes_hopping_and_number_on_same_modes(self):
        self.assertTrue(
            trivially_double_commutes_dual_basis(FermionOperator('4^ 3'),
                                                 FermionOperator('4^ 1'),
                                                 FermionOperator('4^ 1^ 4 1')))

    def test_trivially_double_commutes_no_intersection_a_with_bc(self):
        self.assertTrue(
            trivially_double_commutes_dual_basis(FermionOperator('5^ 2'),
                                                 FermionOperator('3^ 1'),
                                                 FermionOperator('4^ 1^ 4 1')))

    def test_trivially_double_commutes_double_create_in_a_and_b(self):
        self.assertTrue(
            trivially_double_commutes_dual_basis(FermionOperator('5^ 2'),
                                                 FermionOperator('3^ 1'),
                                                 FermionOperator('4^ 1^ 4 1')))

    def test_trivially_double_commutes_double_annihilate_in_a_and_c(self):
        self.assertTrue(
            trivially_double_commutes_dual_basis(FermionOperator('5^ 2'),
                                                 FermionOperator('3^ 1'),
                                                 FermionOperator('4^ 1^ 4 1')))

    def test_no_trivial_double_commute_double_annihilate_with_create(self):
        self.assertFalse(
            trivially_double_commutes_dual_basis(FermionOperator('5^ 2'),
                                                 FermionOperator('2^ 1'),
                                                 FermionOperator('4^ 2')))

    def test_trivially_double_commutes_excess_create(self):
        self.assertTrue(
            trivially_double_commutes_dual_basis(FermionOperator('5^ 2'),
                                                 FermionOperator('5^ 5'),
                                                 FermionOperator('5^ 1')))

    def test_trivially_double_commutes_excess_annihilate(self):
        self.assertTrue(
            trivially_double_commutes_dual_basis(FermionOperator('5^ 2'),
                                                 FermionOperator('3^ 2'),
                                                 FermionOperator('2^ 2')))



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from typing import TYPE_CHECKING

import itertools
from typing import Optional

from openfermion import ops
from openfermion.testing.testing_utils import random_interaction_operator

if TYPE_CHECKING:
    import openfermion


def random_interaction_operator_term(
        order: int,
        real: bool = True,
        seed: Optional[int] = None,
) -> 'openfermion.InteractionOperator':
    """Generates a random interaction operator with non-zero coefficients only
    on terms corresponding to the given number of unique orbitals.

    The number of orbitals is equal to the given order.

    Args:
        order: How many unique orbitals the non-zero terms should correspond to.
        real: Whether or not the coefficients should be real. Defaults to True.
        seed: The seed. If None (default), uses np.random.
    """

    n_orbitals = order

    if order > 4:
        return ops.InteractionOperator.zero(order)

    operator = random_interaction_operator(n_orbitals, real=real, seed=seed)
    operator.constant = 0

    for indices in itertools.product(range(n_orbitals), repeat=2):
        if len(set(indices)) != order:
            operator.one_body_tensor[indices] = 0

    for indices in itertools.product(range(n_orbitals), repeat=4):
        if len(set(indices)) != order:
            operator.two_body_tensor[indices] = 0

    return operator



from typing import Iterable, Generator, Optional, Union
from itertools import zip_longest
import numpy as np
from openfermion.third_party.representability._bijections import Bijection, \
    index_index_basis


class Tensor(object):
    """
    Instantiation of named tensor
    """

    def __init__(self,
                 *,
                 tensor: Optional[Union[None, np.ndarray]] = None,
                 basis: Optional[Union[None, Bijection]] = None,
                 name: Optional[Union[None, str]] = None):
        """
        Named Tensor that allows one to label elements with different indices

        For example, a 2-pdm is labeled by geminals (i, j), (k, l) in matrix
        form.  The matrix representing the 2-RDM can now be indexed into with
        tensor notation (i, j, k, l) instead of calling
        [i * dim + j, l * dim + k]

        Args:
            tensor: numpy.ndarray to hold the tensor data
            basis: basis on all the axes
            name: name of the tensor
        """
        if tensor is not None:
            self.dim = tensor.shape[0]
            self.ndim = tensor.ndim
            self.data = np.copy(tensor)
            self.size = self.dim**self.ndim
            self.name = name
            if basis is None:
                self.basis = index_index_basis(self.dim)
            else:
                if not isinstance(basis, Bijection):
                    raise TypeError("Basis must be a Bijection object")
                self.basis = basis
        else:
            self.dim = None
            self.ndim = None
            self.data = None
            self.size = None
            self.basis = basis
            self.name = name

    def __getitem__(self, indices):
        """
        returns the tensor data if loaded
        """
        if self.data is not None:
            return self.data[indices]
        else:
            raise TypeError("data store is not set")

    def __call__(self, *indices):
        """
        Index into the the data by passing through the basis first

        :param indices: indices for the rev_bas
        :return: element of the data
        """
        # I need a way to find out the dimension of an element of the codomain
        codomain_element_size = self.basis.domain_element_sizes()[1]
        index_set = []
        for idx_set in grouper(indices, codomain_element_size):
            if len(idx_set) == 1:
                index_set.append(self.basis.rev(idx_set[0]))
            else:
                index_set.append(self.basis.rev(idx_set))

        return self.data[tuple(index_set)]

    @staticmethod
    def get_obj_size(obj):
        """
        Determine the number of 'elements' an object contains.

        Integers are 1, tuples and lists are len(tuple/list)

        Args:
            obj: object to query for length
        Returns:  length of the object
        """
        if isinstance(obj, (tuple, list)):
            return len(obj)
        elif isinstance(obj, (float, int, complex, bool)):
            return 1
        else:
            raise TypeError("object type doesn't have a recognized length")

    def index_vectorized(self, *indices):
        """
        Perform the canonical index bijection to a scalar

        Note: the start returns a tuple of n-indices. That includes 1
        """
        return self.index_bijection(self.index_transform(indices), self.ndim,
                                    self.dim)

    def index_transform(self, indices):
        """
        Transform the indices to the basis indices

        :param indices: Tuple of tensor indices
        :return:
        """
        codomain_element_size = self.basis.domain_element_sizes()[1]
        index_set = []
        for idx_set in grouper(indices, codomain_element_size):
            index_set.append(
                self.basis.rev(idx_set[0] if len(idx_set) == 1 else idx_set))

        return tuple(index_set)

    @staticmethod
    def index_bijection(indices, ndim, dim):
        """
        calculate the bijection with tensor dim counting
        """
        if len(indices) != ndim:
            raise TypeError(
                "indices are inappriopriate length for the given ndim")

        # C-order canonical vectorization--i.e. right most index in indices
        # changes with the highest frequency
        bijection = 0
        for n in range(ndim):
            bijection += indices[n] * dim**(ndim - n - 1)
        return bijection

    def utri_iterator(self) -> Generator:
        """
        Iterate over the upper triangle (including diagonal)
        and return data value and index
        """
        return self._iterator("upper")

    def ltri_iterator(self) -> Generator:
        """
        Iterate over the lower triangle (including diagonal)
        and return data value and index
        """
        return self._iterator("lower")

    def all_iterator(self) -> Generator:
        """
        Iterate over the lower triangle (including diagonal)
        and return data value and index
        """
        return self._iterator("all")

    def _iterator(self, ultri: str) -> Generator:
        """
        Iterate over the a data store yielding the upper/lower/all values
        """
        if ultri not in ['upper', 'lower', 'all']:
            raise TypeError(
                "iteration type {} is not 'upper', 'lower', or 'all'".format(
                    ultri))

        it = np.nditer(self.data, flags=['multi_index'])
        while not it.finished:
            indices = it.multi_index
            left_idx_set = self.index_bijection(indices[:self.ndim // 2],
                                                self.ndim // 2, self.dim)
            right_idx_set = self.index_bijection(indices[self.ndim // 2:],
                                                 self.ndim // 2, self.dim)

            if ultri == 'upper' and left_idx_set <= right_idx_set:
                yield it[0], map(lambda x: self.basis.fwd(x), it.multi_index)
            elif ultri == 'lower' and left_idx_set >= right_idx_set:
                yield it[0], map(lambda x: self.basis.fwd(x), it.multi_index)
            elif ultri == 'all':
                yield it[0], map(lambda x: self.basis.fwd(x), it.multi_index)

            it.iternext()

    def vectorize(self, order: Optional[str] = 'C') -> np.ndarray:
        """
        Take a multidimensional array and vectorized via C ordering

        :return: a vector of self.size x 1
        """
        return np.reshape(self.data, (-1, 1), order=order)


# from standard library itertools recipe book
def grouper(iterable: Iterable,
            n: int,
            fillvalue: Optional[Union[None, str]] = None):
    """Collect data into fixed-length chunks or blocks"""
    # grouper('ABCDEFG', 3, 'x') --> ABC DEF Gxx"
    args = [iter(iterable)] * n
    return zip_longest(*args, fillvalue=fillvalue)



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Tests for linear_qubit_operator.py."""

import multiprocessing
import unittest

import numpy
import scipy.sparse.linalg

from openfermion.ops.operators import QubitOperator
from openfermion.linalg.linear_qubit_operator import (
    LinearQubitOperator,
    LinearQubitOperatorOptions,
    ParallelLinearQubitOperator,
    apply_operator,
    generate_linear_qubit_operator,
)
from openfermion.linalg.sparse_tools import qubit_operator_sparse


class LinearQubitOperatorOptionsTest(unittest.TestCase):
    """Tests for LinearQubitOperatorOptions class."""

    def setUp(self):
        """LinearQubitOperatorOptions test set up."""
        self.processes = multiprocessing.cpu_count()
        self.options = LinearQubitOperatorOptions(self.processes)

    def test_init(self):
        """Tests __init__()."""
        self.assertEqual(self.options.processes, self.processes)
        self.assertIsNone(self.options.pool)

    def test_get_processes_small(self):
        """Tests get_processes() with a small num."""
        num = 1
        self.assertEqual(self.options.get_processes(num), num)

    def test_get_processes_large(self):
        """Tests get_processes() with a large num."""
        self.assertEqual(self.options.get_processes(2 * self.processes),
                         self.processes)

    def test_invalid_processes(self):
        """Tests with invalid processes since it's not positive."""
        with self.assertRaises(ValueError):
            LinearQubitOperatorOptions(0)

    def test_get_pool(self):
        """Tests get_pool() without a num."""
        self.assertIsNone(self.options.pool)

        pool = self.options.get_pool()
        self.assertIsNotNone(pool)

    def test_get_pool_with_num(self):
        """Tests get_processes() with a num."""
        self.assertIsNone(self.options.pool)

        pool = self.options.get_pool(2)
        self.assertIsNotNone(pool)


class LinearQubitOperatorTest(unittest.TestCase):
    """Tests for LinearQubitOperator class."""

    def test_init(self):
        """Tests __init__()."""
        qubit_operator = QubitOperator('Z2')
        n_qubits = 3
        linear_operator = LinearQubitOperator(qubit_operator)

        self.assertEqual(linear_operator.qubit_operator, qubit_operator)
        self.assertEqual(linear_operator.n_qubits, n_qubits)

        # Checks type.
        self.assertTrue(
            isinstance(linear_operator, scipy.sparse.linalg.LinearOperator))

    def test_matvec_wrong_n(self):
        """Testing with wrong n_qubits."""
        with self.assertRaises(ValueError):
            LinearQubitOperator(QubitOperator('X3'), 1)

    def test_matvec_wrong_vec_length(self):
        """Testing with wrong vector length."""
        with self.assertRaises(ValueError):
            _ = LinearQubitOperator(QubitOperator('X3')) * numpy.zeros(4)

    def test_matvec_0(self):
        """Testing with zero term."""
        qubit_operator = QubitOperator.zero()

        vec = numpy.array([1, 2, 3, 4, 5, 6, 7, 8])
        matvec_expected = numpy.zeros(vec.shape)

        self.assertTrue(
            numpy.allclose(
                LinearQubitOperator(qubit_operator, 3) * vec, matvec_expected))

    def test_matvec_x(self):
        """Testing product with X."""
        vec = numpy.array([1, 2, 3, 4])
        matvec_expected = numpy.array([2, 1, 4, 3])

        self.assertTrue(
            numpy.allclose(
                LinearQubitOperator(QubitOperator('X1')) * vec,
                matvec_expected))

    def test_matvec_y(self):
        """Testing product with Y."""
        vec = numpy.array([1, 2, 3, 4], dtype=complex)
        matvec_expected = 1.0j * numpy.array([-2, 1, -4, 3], dtype=complex)

        self.assertTrue(
            numpy.allclose(
                LinearQubitOperator(QubitOperator('Y1')) * vec,
                matvec_expected))

    def test_matvec_z(self):
        """Testing product with Z."""
        vec = numpy.array([1, 2, 3, 4])
        matvec_expected = numpy.array([1, 2, -3, -4])

        self.assertTrue(
            numpy.allclose(
                LinearQubitOperator(QubitOperator('Z0'), 2) * vec,
                matvec_expected))

    def test_matvec_z3(self):
        """Testing product with Z^n."""
        vec = numpy.array(
            [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16])
        matvec_expected = numpy.array(
            [1, -2, 3, -4, 5, -6, 7, -8, 9, -10, 11, -12, 13, -14, 15, -16])

        self.assertTrue(
            numpy.allclose(
                LinearQubitOperator(QubitOperator('Z3')) * vec,
                matvec_expected))

    def test_matvec_zx(self):
        """Testing with multiple factors."""
        vec = numpy.array([1, 2, 3, 4])
        matvec_expected = numpy.array([2, 1, -4, -3])

        self.assertTrue(
            numpy.allclose(
                LinearQubitOperator(QubitOperator('Z0 X1')) * vec,
                matvec_expected))

    def test_matvec_multiple_terms(self):
        """Testing with multiple terms."""
        qubit_operator = (QubitOperator.identity() + 2 * QubitOperator('Y2') +
                          QubitOperator(((0, 'Z'), (1, 'X')), 10.0))

        vec = numpy.array([1, 2, 3, 4, 5, 6, 7, 8])
        matvec_expected = (10 * numpy.array([3, 4, 1, 2, -7, -8, -5, -6]) +
                           2j * numpy.array([-2, 1, -4, 3, -6, 5, -8, 7]) + vec)

        self.assertTrue(
            numpy.allclose(
                LinearQubitOperator(qubit_operator) * vec, matvec_expected))

    def test_matvec_compare(self):
        """Compare LinearQubitOperator with qubit_operator_sparse."""
        qubit_operator = QubitOperator('X0 Y1 Z3')
        mat_expected = qubit_operator_sparse(qubit_operator)

        self.assertTrue(
            numpy.allclose(
                numpy.transpose(
                    numpy.array([
                        LinearQubitOperator(qubit_operator) * v
                        for v in numpy.identity(16)
                    ])), mat_expected.A))


class ParallelLinearQubitOperatorTest(unittest.TestCase):
    """Tests for ParallelLinearQubitOperator class."""

    def setUp(self):
        """ParallelLinearQubitOperator test set up."""
        self.qubit_operator = (QubitOperator('Z3') + QubitOperator('Y0') +
                               QubitOperator('X1'))
        self.n_qubits = 4
        self.linear_operator = ParallelLinearQubitOperator(self.qubit_operator)

        # Vectors for calculations.
        self.vec = numpy.array(range(2**self.n_qubits))

        expected_matvec = numpy.array([
            0,
            -1,
            2,
            -3,
            4,
            -5,
            6,
            -7,
            8,
            -9,
            10,
            -11,
            12,
            -13,
            14,
            -15,
        ])
        expected_matvec = expected_matvec + numpy.array([
            -8j,
            -9j,
            -10j,
            -11j,
            -12j,
            -13j,
            -14j,
            -15j,
            0j,
            1j,
            2j,
            3j,
            4j,
            5j,
            6j,
            7j,
        ])
        expected_matvec += numpy.array([
            4,
            5,
            6,
            7,
            0,
            1,
            2,
            3,
            12,
            13,
            14,
            15,
            8,
            9,
            10,
            11,
        ])
        self.expected_matvec = expected_matvec

    def test_init(self):
        """Tests __init__()."""
        self.assertEqual(self.linear_operator.qubit_operator,
                         self.qubit_operator)
        self.assertEqual(self.linear_operator.n_qubits, self.n_qubits)
        self.assertIsNone(self.linear_operator.options.pool)

        cpu_count = multiprocessing.cpu_count()
        default_processes = min(cpu_count, 10)
        self.assertEqual(self.linear_operator.options.processes,
                         default_processes)

        # Generated variables.
        self.assertEqual(len(self.linear_operator.qubit_operator_groups),
                         min(multiprocessing.cpu_count(), 3))
        self.assertEqual(
            QubitOperator.accumulate(
                self.linear_operator.qubit_operator_groups),
            self.qubit_operator)

        for linear_operator in self.linear_operator.linear_operators:
            self.assertEqual(linear_operator.n_qubits, self.n_qubits)
            self.assertTrue(isinstance(linear_operator, LinearQubitOperator))

        # Checks type.
        self.assertTrue(
            isinstance(self.linear_operator,
                       scipy.sparse.linalg.LinearOperator))

    def test_matvec(self):
        """Tests _matvec() for matrix multiplication with a vector."""

        self.assertIsNone(self.linear_operator.options.pool)
        self.assertTrue(
            numpy.allclose(self.linear_operator * self.vec,
                           self.expected_matvec))

    def test_matvec_0(self):
        """Testing with zero term."""
        qubit_operator = QubitOperator.zero()

        vec = numpy.array([1, 2, 3, 4, 5, 6, 7, 8])
        matvec_expected = numpy.zeros(vec.shape)

        self.assertTrue(
            numpy.allclose(
                ParallelLinearQubitOperator(qubit_operator, 3) * vec,
                matvec_expected))
        self.assertIsNone(self.linear_operator.options.pool)

    def test_closed_workers_not_reused(self):
        qubit_operator = QubitOperator('X0')
        parallel_qubit_op = ParallelLinearQubitOperator(
            qubit_operator, 1, options=LinearQubitOperatorOptions(processes=2))
        state = [1.0, 0.0]
        parallel_qubit_op.dot(state)
        parallel_qubit_op.dot(state)
        self.assertIsNone(parallel_qubit_op.options.pool)


class UtilityFunctionTest(unittest.TestCase):
    """Tests for utility functions."""

    def test_apply_operator(self):
        """Tests apply_operator() since it's executed on other processors."""
        vec = numpy.array([1, 2, 3, 4])
        matvec_expected = numpy.array([2, 1, 4, 3])

        self.assertTrue(
            numpy.allclose(
                apply_operator((LinearQubitOperator(QubitOperator('X1')), vec)),
                matvec_expected))

    def test_generate_linear_operator(self):
        """Tests generate_linear_qubit_operator()."""
        qubit_operator = (QubitOperator('Z3') + QubitOperator('X1') +
                          QubitOperator('Y0'))
        n_qubits = 6

        # Checks types.
        operator = generate_linear_qubit_operator(qubit_operator, n_qubits)
        self.assertTrue(isinstance(operator, LinearQubitOperator))
        self.assertFalse(isinstance(operator, ParallelLinearQubitOperator))

        operator_again = generate_linear_qubit_operator(
            qubit_operator, n_qubits, options=LinearQubitOperatorOptions(2))
        self.assertTrue(isinstance(operator_again, ParallelLinearQubitOperator))
        self.assertFalse(isinstance(operator_again, LinearQubitOperator))

        # Checks operators are equivalent.
        numpy.random.seed(n_qubits)
        vec = numpy.random.rand(2**n_qubits, 1)
        self.assertTrue(numpy.allclose(operator * vec, operator_again * vec))



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Module to manipulate basic models of quantum channels"""

from functools import reduce
from itertools import chain
from numpy import array, dot, eye, kron, log2, sqrt


def _verify_channel_inputs(density_matrix, probability, target_qubit):
    r"""Verifies input parameters for channels

    Args:
        density_matrix (numpy.ndarray): Density matrix of the system
        probability (float): Probability error is applied p \in [0, 1]
        target_qubit (int): target for the channel error.

    Returns:
        new_density_matrix(numpy.ndarray): Density matrix with the channel
            applied.
    """
    n_qubits = int(log2(density_matrix.shape[0]))

    if (len(density_matrix.shape) != 2 or
            density_matrix.shape[0] != density_matrix.shape[1]):
        raise ValueError("Error in input of density matrix to channel.")
    if (probability < 0) or (probability > 1):
        raise ValueError("Channel probability must be between 0 and 1.")
    if (target_qubit < 0) or (target_qubit >= n_qubits):
        raise ValueError("Target qubits must be within number of qubits.")


def _lift_operator(operator, n_qubits, target_qubit):
    """Lift a single qubit operator into the n_qubit space by kron product

    Args:
        operator (ndarray): Single qubit operator to lift into full space
        n_qubits (int): Number of total qubits in the space
        target_qubit (int): Qubit to act on

    Return:
        new_operator(Sparse Operator): Operator representing the embedding in
            the full space.
    """
    new_operator = (reduce(
        kron,
        chain((eye(2) for i in range(0, target_qubit)), [operator],
              (eye(2) for i in range(target_qubit + 1, n_qubits)))))
    return new_operator


def amplitude_damping_channel(density_matrix,
                              probability,
                              target_qubit,
                              transpose=False):
    r"""Apply an amplitude damping channel

    Applies an amplitude damping channel with a given probability to the target
    qubit in the density_matrix.

    Args:
        density_matrix (numpy.ndarray): Density matrix of the system
        probability (float): Probability error is applied p \in [0, 1]
        target_qubit (int): target for the channel error.
        transpose (bool): Conjugate transpose channel operators, useful for
            acting on Hamiltonians in variational channel state models

    Returns:
        new_density_matrix(numpy.ndarray): Density matrix with the channel
            applied.
    """
    _verify_channel_inputs(density_matrix, probability, target_qubit)
    n_qubits = int(log2(density_matrix.shape[0]))

    E0 = _lift_operator(
        array([[1.0, 0.0], [0.0, sqrt(1.0 - probability)]], dtype=complex),
        n_qubits, target_qubit)

    E1 = _lift_operator(
        array([[0.0, sqrt(probability)], [0.0, 0.0]], dtype=complex), n_qubits,
        target_qubit)

    if transpose:
        E0 = E0.T
        E1 = E1.T

    new_density_matrix = (dot(E0, dot(density_matrix, E0.T)) +
                          dot(E1, dot(density_matrix, E1.T)))

    return new_density_matrix


def dephasing_channel(density_matrix,
                      probability,
                      target_qubit,
                      transpose=False):
    r"""Apply a dephasing channel

    Applies an amplitude damping channel with a given probability to the target
    qubit in the density_matrix.

    Args:
        density_matrix (numpy.ndarray): Density matrix of the system
        probability (float): Probability error is applied p \in [0, 1]
        target_qubit (int): target for the channel error.
        transpose (bool): Conjugate transpose channel operators, useful for
            acting on Hamiltonians in variational channel state models

    Returns:
        new_density_matrix (numpy.ndarray): Density matrix with the channel
            applied.
    """
    _verify_channel_inputs(density_matrix, probability, target_qubit)
    n_qubits = int(log2(density_matrix.shape[0]))

    E0 = _lift_operator(
        sqrt(1.0 - probability / 2.) * eye(2), n_qubits, target_qubit)
    E1 = _lift_operator(
        sqrt(probability / 2.) * array([[1.0, 0.0], [1.0, -1.0]]), n_qubits,
        target_qubit)

    if transpose:
        E0 = E0.T
        E1 = E1.T

    new_density_matrix = (dot(E0, dot(density_matrix, E0.T)) +
                          dot(E1, dot(density_matrix, E1.T)))

    return new_density_matrix


def depolarizing_channel(density_matrix,
                         probability,
                         target_qubit,
                         transpose=False):
    r"""Apply a depolarizing channel

    Applies an amplitude damping channel with a given probability to the target
    qubit in the density_matrix.

    Args:
        density_matrix (numpy.ndarray): Density matrix of the system
        probability (float): Probability error is applied p \in [0, 1]
        target_qubit (int/str): target for the channel error, if given special
            value "all", then a total depolarizing channel is applied.
        transpose (bool): Dummy parameter to match signature of other
            channels but depolarizing channel is symmetric under
            conjugate transpose.

    Returns:
        new_density_matrix (numpy.ndarray): Density matrix with the channel
            applied.
    """
    n_qubits = int(log2(density_matrix.shape[0]))

    # Toggle depolarizing channel on all qubits
    if isinstance(target_qubit, str) and target_qubit.lower() == "all":
        dimension = density_matrix.shape[0]
        new_density_matrix = ((1.0 - probability) * density_matrix +
                              probability * eye(dimension) / float(dimension))
        return new_density_matrix

    # For any other case, depolarize only the target qubit
    _verify_channel_inputs(density_matrix, probability, target_qubit)

    E0 = _lift_operator(
        sqrt(1.0 - probability) * eye(2), n_qubits, target_qubit)
    E1 = _lift_operator(
        sqrt(probability / 3.) * array([[0.0, 1.0], [1.0, 0.0]]), n_qubits,
        target_qubit)
    E2 = _lift_operator(
        sqrt(probability / 3.) * array([[0.0, -1.0j], [1.0j, 0.0]]), n_qubits,
        target_qubit)
    E3 = _lift_operator(
        sqrt(probability / 3.) * array([[1.0, 0.0], [0.0, -1.0]]), n_qubits,
        target_qubit)

    new_density_matrix = (dot(E0, dot(density_matrix, E0)) +
                          dot(E1, dot(density_matrix, E1)) +
                          dot(E2, dot(density_matrix, E2)) +
                          dot(E3, dot(density_matrix, E3)))

    return new_density_matrix



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import pytest

from openfermion import FermionOperator
from openfermion.circuits.trotter import TrotterStep


def test_trotter_algorithm_is_abstract_cant_instantiate():
    with pytest.raises(TypeError):
        _ = TrotterStep(FermionOperator())


def test_trotter_algorithm_is_abstract_must_implement():

    class Missing(TrotterStep):
        pass

    with pytest.raises(TypeError):
        _ = Missing(FermionOperator())


def test_trotter_algorithm_is_abstract_can_implement():

    class Included(TrotterStep):

        def trotter_step(self, qubits, time, control_qubit):
            pass

    assert isinstance(Included(FermionOperator()), TrotterStep)



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
'''Functions for fitting simple oscillating functions'''

from typing import Tuple
import numpy
import scipy


def fit_known_frequencies(signal: numpy.ndarray, times: numpy.ndarray,
                          frequencies: numpy.ndarray) -> numpy.ndarray:
    """Fits a set of known exponential components to a dataset

    Decomposes a function g(t) as g(t)=sum_jA_jexp(iw_jt), where the frequencies
    w_j are already known. Namely, makes a least-squares fit.

    Arguments:
        signal {numpy.ndarray} -- the signal g(t) to be fit
        times {numpy.ndarray} -- t values of the signal
        frequencies {numpy.ndarray} -- known frequencies w_j

    Returns:
        amplitudes {numpy.ndarray} -- the found amplitudes A_j
    """
    generation_matrix = numpy.array([
        [numpy.exp(1j * time * freq) for freq in frequencies] for time in times
    ])
    amplitudes = scipy.linalg.lstsq(generation_matrix, signal)[0]
    return amplitudes


def prony(signal: numpy.ndarray) -> Tuple[numpy.ndarray, numpy.ndarray]:
    """Estimates amplitudes and phases of a sparse signal using Prony's method.

    Single-ancilla quantum phase estimation returns a signal
    g(k)=sum (aj*exp(i*k*phij)), where aj and phij are the amplitudes
    and corresponding eigenvalues of the unitary whose phases we wish
    to estimate. When more than one amplitude is involved, Prony's method
    provides a simple estimation tool, which achieves near-Heisenberg-limited
    scaling (error scaling as N^{-1/2}K^{-3/2}).

    Args:
        signal(1d complex array): the signal to fit

    Returns:
        amplitudes(list of complex values): the amplitudes a_i,
        in descending order by their complex magnitude
        phases(list of complex values): the complex frequencies gamma_i,
            correlated with amplitudes.
    """

    num_freqs = len(signal) // 2
    hankel0 = scipy.linalg.hankel(c=signal[:num_freqs],
                                  r=signal[num_freqs - 1:-1])
    hankel1 = scipy.linalg.hankel(c=signal[1:num_freqs + 1],
                                  r=signal[num_freqs:])
    shift_matrix = scipy.linalg.lstsq(hankel0.T, hankel1.T)[0]
    phases = numpy.linalg.eigvals(shift_matrix.T)

    generation_matrix = numpy.array(
        [[phase**k for phase in phases] for k in range(len(signal))])
    amplitudes = scipy.linalg.lstsq(generation_matrix, signal)[0]

    amplitudes, phases = zip(*sorted(
        zip(amplitudes, phases), key=lambda x: numpy.abs(x[0]), reverse=True))

    return numpy.array(amplitudes), numpy.array(phases)



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
'''Functions to reorder terms within SymbolicOperators'''

import itertools
import numpy

from openfermion.ops.operators import (BosonOperator, FermionOperator,
                                       QuadOperator)
from openfermion.ops.representations import InteractionOperator


def chemist_ordered(fermion_operator):
    r"""Puts a two-body fermion operator in chemist ordering.

    The normal ordering convention for chemists is different.
    Rather than ordering the two-body term as physicists do, as
    :math:`a^\dagger a^\dagger a a`
    the chemist ordering of the two-body term is
    :math:`a^\dagger a a^\dagger a`

    TODO: This routine can be made more efficient.

    Args:
        fermion_operator (FermionOperator): a fermion operator guarenteed to
            have number conserving one- and two-body fermion terms only.
    Returns:
        chemist_ordered_operator (FermionOperator): the input operator
            ordered in the chemistry convention.
    Raises:
        OperatorSpecificationError: Operator is not two-body number conserving.
    """
    # Make sure we're dealing with a fermion operator from a molecule.
    if not fermion_operator.is_two_body_number_conserving():
        raise TypeError('Operator is not two-body number conserving.')

    # Normal order and begin looping.
    normal_ordered_input = normal_ordered(fermion_operator)
    chemist_ordered_operator = FermionOperator()
    for term, coefficient in normal_ordered_input.terms.items():
        if len(term) == 2 or not len(term):
            chemist_ordered_operator += FermionOperator(term, coefficient)
        else:
            # Possibly add new one-body term.
            if term[1][0] == term[2][0]:
                new_one_body_term = (term[0], term[3])
                chemist_ordered_operator += FermionOperator(
                    new_one_body_term, coefficient)
            # Reorder two-body term.
            new_two_body_term = (term[0], term[2], term[1], term[3])
            chemist_ordered_operator += FermionOperator(new_two_body_term,
                                                        -coefficient)
    return chemist_ordered_operator


def normal_ordered(operator, hbar=1.):
    r"""Compute and return the normal ordered form of a FermionOperator,
    BosonOperator, QuadOperator, or InteractionOperator.

    Due to the canonical commutation/anticommutation relations satisfied
    by these operators, there are multiple forms that the same operator
    can take. Here, we define the normal ordered form of each operator,
    providing a distinct representation for distinct operators.

    In our convention, normal ordering implies terms are ordered
    from highest tensor factor (on left) to lowest (on right). In
    addition:

    * FermionOperators: a^\dagger comes before a
    * BosonOperators: b^\dagger comes before b
    * QuadOperators: q operators come before p operators,

    Args:
        operator: an instance of the FermionOperator, BosonOperator,
            QuadOperator, or InteractionOperator classes.
        hbar (float): the value of hbar used in the definition of the
            commutator [q_i, p_j] = i hbar delta_ij. By default hbar=1.
            This argument only applies when normal ordering QuadOperators.
    """
    kwargs = {}

    if isinstance(operator, FermionOperator):
        ordered_operator = FermionOperator()
        order_fn = normal_ordered_ladder_term
        kwargs['parity'] = -1

    elif isinstance(operator, BosonOperator):
        ordered_operator = BosonOperator()
        order_fn = normal_ordered_ladder_term
        kwargs['parity'] = 1

    elif isinstance(operator, QuadOperator):
        ordered_operator = QuadOperator()
        order_fn = normal_ordered_quad_term
        kwargs['hbar'] = hbar

    elif isinstance(operator, InteractionOperator):
        constant = operator.constant
        n_modes = operator.n_qubits
        one_body_tensor = operator.one_body_tensor.copy()
        two_body_tensor = numpy.zeros_like(operator.two_body_tensor)
        quadratic_index_pairs = (
            (pq, pq) for pq in itertools.combinations(range(n_modes)[::-1], 2))
        cubic_index_pairs = (
            index_pair
            for p, q, r in itertools.combinations(range(n_modes)[::-1], 3)
            for index_pair in [((p, q), (p, r)), ((p, r), (
                p, q)), ((p, q), (q, r)), ((q, r),
                                           (p, q)), ((p, r),
                                                     (q, r)), ((q, r), (p, r))])
        quartic_index_pairs = (
            index_pair
            for p, q, r, s in itertools.combinations(range(n_modes)[::-1], 4)
            for index_pair in [((p, q), (r, s)), ((r, s), (
                p, q)), ((p, r), (q, s)), ((q, s),
                                           (p, r)), ((p, s),
                                                     (q, r)), ((q, r), (p, s))])
        index_pairs = itertools.chain(quadratic_index_pairs, cubic_index_pairs,
                                      quartic_index_pairs)
        for pq, rs in index_pairs:
            two_body_tensor[pq + rs] = sum(
                s * ss * operator.two_body_tensor[pq[::s] + rs[::ss]]
                for s, ss in itertools.product([-1, 1], repeat=2))
        return InteractionOperator(constant, one_body_tensor, two_body_tensor)

    else:
        raise TypeError('Can only normal order FermionOperator, '
                        'BosonOperator, QuadOperator, or InteractionOperator.')

    for term, coefficient in operator.terms.items():
        ordered_operator += order_fn(term, coefficient, **kwargs)

    return ordered_operator


def normal_ordered_ladder_term(term, coefficient, parity=-1):
    """Return a normal ordered FermionOperator or BosonOperator corresponding
    to single term.

    Args:
        term (list or tuple): A sequence of tuples. The first element of each
            tuple is an integer indicating the mode on which a fermion ladder
            operator acts, starting from zero. The second element of each
            tuple is an integer, either 1 or 0, indicating whether creation
            or annihilation acts on that mode.
        coefficient(complex or float): The coefficient of the term.
        parity (int): parity=-1 corresponds to a Fermionic term that should be
            ordered based on the canonical anti-commutation relations.
            parity=1 corresponds to a Bosonic term that should be ordered based
            on the canonical commutation relations.

    Returns:
        ordered_term: a FermionOperator or BosonOperator instance.
            The normal ordered form of the input.
            Note that this might have more terms.

    In our convention, normal ordering implies terms are ordered
    from highest tensor factor (on left) to lowest (on right).
    Also, ladder operators come first.

    Warning:
        Even assuming that each creation or annihilation operator appears
        at most a constant number of times in the original term, the
        runtime of this method is exponential in the number of qubits.
    """
    # Iterate from left to right across operators and reorder to normal
    # form. Swap terms operators into correct position by moving from
    # left to right across ladder operators.
    term = list(term)

    if parity == -1:
        Op = FermionOperator
    elif parity == 1:
        Op = BosonOperator

    ordered_term = Op()

    for i in range(1, len(term)):
        for j in range(i, 0, -1):
            right_operator = term[j]
            left_operator = term[j - 1]

            # Swap operators if raising on right and lowering on left.
            if right_operator[1] and not left_operator[1]:
                term[j - 1] = right_operator
                term[j] = left_operator
                coefficient *= parity

                # Replace a a^\dagger with 1 + parity*a^\dagger a
                # if indices are the same.
                if right_operator[0] == left_operator[0]:
                    new_term = term[:(j - 1)] + term[(j + 1):]

                    # Recursively add the processed new term.
                    ordered_term += normal_ordered_ladder_term(
                        tuple(new_term), parity * coefficient, parity)

            # Handle case when operator type is the same.
            elif right_operator[1] == left_operator[1]:

                # If same two Fermionic operators are repeated,
                # evaluate to zero.
                if parity == -1 and right_operator[0] == left_operator[0]:
                    return ordered_term

                # Swap if same ladder type but lower index on left.
                elif right_operator[0] > left_operator[0]:
                    term[j - 1] = right_operator
                    term[j] = left_operator
                    coefficient *= parity

    # Add processed term and return.
    ordered_term += Op(tuple(term), coefficient)
    return ordered_term


def normal_ordered_quad_term(term, coefficient, hbar=1.):
    """Return a normal ordered QuadOperator corresponding to single term.

    Args:
        term: A tuple of tuples. The first element of each tuple is
            an integer indicating the mode on which a boson ladder
            operator acts, starting from zero. The second element of each
            tuple is an integer, either 1 or 0, indicating whether creation
            or annihilation acts on that mode.
        coefficient: The coefficient of the term.
        hbar (float): the value of hbar used in the definition of the
            commutator [q_i, p_j] = i hbar delta_ij. By default hbar=1.

    Returns:
        ordered_term (QuadOperator): The normal ordered form of the input.
            Note that this might have more terms.

    In our convention, normal ordering implies terms are ordered
    from highest tensor factor (on left) to lowest (on right).
    Also, q operators come first.
    """
    # Iterate from left to right across operators and reorder to normal
    # form. Swap terms operators into correct position by moving from
    # left to right across ladder operators.
    term = list(term)
    ordered_term = QuadOperator()
    for i in range(1, len(term)):
        for j in range(i, 0, -1):
            right_operator = term[j]
            left_operator = term[j - 1]

            # Swap operators if q on right and p on left.
            # p q -> q p
            if right_operator[1] == 'q' and not left_operator[1] == 'q':
                term[j - 1] = right_operator
                term[j] = left_operator

                # Replace p q with i hbar + q p
                # if indices are the same.
                if right_operator[0] == left_operator[0]:
                    new_term = term[:(j - 1)] + term[(j + 1)::]

                    # Recursively add the processed new term.
                    ordered_term += normal_ordered_quad_term(
                        tuple(new_term), -coefficient * 1j * hbar)

            # Handle case when operator type is the same.
            elif right_operator[1] == left_operator[1]:

                # Swap if same type but lower index on left.
                if right_operator[0] > left_operator[0]:
                    term[j - 1] = right_operator
                    term[j] = left_operator

    # Add processed term and return.
    ordered_term += QuadOperator(tuple(term), coefficient)
    return ordered_term


def reorder(operator, order_function, num_modes=None, reverse=False):
    """Changes the ladder operator order of the Hamiltonian based on the
    provided order_function per mode index.

    Args:
        operator (SymbolicOperator): the operator that will be reordered. must
            be a SymbolicOperator or any type of operator that inherits from
            SymbolicOperator.
        order_function (func): a function per mode that is used to map the
            indexing. must have arguments mode index and num_modes.
        num_modes (int): default None. User can provide the number of modes
            assumed for the system. if None, the number of modes will be
            calculated based on the Operator.
        reverse (bool): default False. if set to True, the mode mapping is
            reversed. reverse = True will not revert back to original if
            num_modes calculated differs from original and reverted.

    Note: Every order function must take in a mode_idx and num_modes.
    """

    if num_modes is None:
        num_modes = max(
            [factor[0] for term in operator.terms for factor in term]) + 1

    mode_map = {
        mode_idx: order_function(mode_idx, num_modes)
        for mode_idx in range(num_modes)
    }

    if reverse:
        mode_map = {val: key for key, val in mode_map.items()}

    rotated_hamiltonian = operator.__class__()
    for term, value in operator.terms.items():
        new_term = tuple([(mode_map[op[0]], op[1]) for op in term])
        rotated_hamiltonian += operator.__class__(new_term, value)
    return rotated_hamiltonian



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""This module contains methods to lift tensors to higher spaces
through the Grassmann wedge product."""
from itertools import product
from math import factorial
import copy
import numpy

# numpy einsum alphabet
EINSUM_CHARS = 'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ'


def generate_parity_permutations(seq):
    """
    Generates the permutations and sign of a sequence by constructing a tree
    where the nth level contains all n-permutations of m (n < m) objects.

    At the last level where n == m all permutations are generated.  The sign
    is kept updated by determining where the next number is inserted into the
    current leaf's set of numbers.

    Example:
         Constructing the permutations of the sequence [A, B, C] constructs the
         following tree:

         [[(A, +1)], [(AB, +1), (BA, -1)], [(ABC, +1), (ACB, -1), (CAB, +1],
         [(BAC, -1), (BCA, +1), (CBA, -1)]]

    Args:
        seq: a sequence of a string to provide permutations

    returns:
        a permutation list with the elements of the seq permuted and a sign
        associated with the permutation.
    """
    if isinstance(seq, str):
        seq = [x for x in seq]

    indices = seq[1:]
    permutations = [([seq[0]], 1)]
    while indices:
        index_to_inject = indices.pop(0)

        new_permutations = []  # permutations in the tree
        for perm in permutations:
            # now loop over positions to insert
            for put_index in range(len(perm[0]) + 1):
                new_index_list = copy.deepcopy(perm[0])
                # insert new object starting at end of the list
                new_index_list.insert(len(perm[0]) - put_index, index_to_inject)

                new_permutations.append(
                    (new_index_list, perm[1] * (-1)**(put_index)))

        permutations = new_permutations

    return permutations


def wedge(left_tensor, right_tensor, left_index_ranks, right_index_ranks):
    """
    Implement the wedge product between left_tensor and right_tensor

    The wedge product is defined as

    .. math::
        \\begin{align}
        a_{j_{1}, j_{2}, ...,j_{p}}^{i_{1}, i_{2}, ..., i_{p}} \\wedge
        b_{j_{p+1}, j_{p+2}, ..., j_{N}}^{i_{p+1}, i_{p + 2}, ..., i_{N}} =
        \\left(\\frac{1}{N!}\\right)^{2} = \\sum_{\\pi, \\sigma}\\epsilon(\\pi)
        \\epsilon(\\sigma)\\pi \\sigma
        a_{j_{1}, j_{2}, ...,j_{p}}^{i_{1}, i_{2}, ..., i_{p}}
        b_{j_{p+1}, j_{p+2}, ..., j_{N}}^{i_{p+1}, i_{p + 2}, ..., i_{N}}
        \\end{align}

    The top indices are those that transform contravariently.  The bottom
    indices transform covariently.

    The tensor storage convention for marginals follows the OpenFermion
    convention. tpdm[i, j, k, l] = <i^ j^ k l>,
    rtensor[u1, u2, u3, d1] = <u1^ u2^ u3^ d1>

    Args:
        left_tensor: left tensor to wedge product
        right_tensor: right tensor to wedge product
        left_index_ranks: tuple of number of indices that transform
                          contravariently and covariently
        right_index_ranks: tuple of number of indices that transform
                           contravariently and covariently
    Returns:
        new tensor constructed as the wedge product of the left_tensor and
        right_tensor
    """
    if left_tensor.ndim != sum(left_index_ranks):
        raise IndexError(
            "n_tensor shape is not consistent with the input n_index rank")
    if right_tensor.ndim != sum(right_index_ranks):
        raise IndexError(
            "n_tensor shape is not consistent with the input n_index rank")
    # assign upper and lower indices for n_tensor
    total_upper = left_index_ranks[0] + right_index_ranks[0]
    total_lower = left_index_ranks[1] + right_index_ranks[1]
    upper_characters = EINSUM_CHARS[:total_upper]
    lower_characters = EINSUM_CHARS[total_upper:total_upper + total_lower]
    new_tensor = numpy.zeros(left_tensor.shape + right_tensor.shape,
                             dtype=complex)
    ordered_einsum_string = upper_characters + lower_characters

    for upper_order_parities, lower_order_parities in product(
            generate_parity_permutations(upper_characters),
            generate_parity_permutations(lower_characters[::-1])):
        # we reverse the order in the lower_chars so because
        # <a^ b^ c d> = D_{dc}^{ab} in this code.
        n_upper_einsum_chars = upper_order_parities[0][:left_index_ranks[0]]
        m_upper_einsum_chars = upper_order_parities[0][left_index_ranks[0]:]
        n_lower_einsum_chars = lower_order_parities[0] \
            [:left_index_ranks[1]][::-1]
        m_lower_einsum_chars = lower_order_parities[0] \
            [left_index_ranks[1]:][::-1]

        n_string = "".join(n_upper_einsum_chars + n_lower_einsum_chars)
        m_string = "".join(m_upper_einsum_chars + m_lower_einsum_chars)

        # we are doing lots of extra += operations but with the benefit of not
        # having to write a python loop over the entire new_tensor object.
        new_tensor += upper_order_parities[1] * lower_order_parities[1] * \
                      numpy.einsum('{},{}->{}'.format(n_string, m_string,
                                                      ordered_einsum_string),
                                   left_tensor, right_tensor)

    new_tensor /= factorial(total_upper) * factorial(total_lower)

    return new_tensor



from typing import Callable, List, Tuple


class Bijection:

    def __init__(self, fwd: Callable, rev: Callable, sizes: Callable):
        """
        Bijection holds forward maps and backwards maps

        Args:
            fwd: forward mapping function
            rev: backwards mapping function
            sizes: function getting tuple of domain and codomain element
                   dimensions
        """
        self.fwd = fwd
        self.rev = rev
        self.domain_element_sizes = sizes


def index_index_basis(n: int) -> Bijection:
    """
    Create an index-to-index Bijection

    Args:
        n: length of basis
    Returns: A Bijection object
    """
    idx_dict = dict(zip(range(n), range(n)))

    def forward(i: int) -> int:
        return idx_dict[i]

    def reverse(i: int) -> int:
        return idx_dict[i]

    def sizes() -> Tuple[int, int]:
        return 1, 1

    return Bijection(forward, reverse, sizes)


def index_tuple_basis(codomain_elements: List[Tuple[int, ...]]) -> Bijection:
    """
    Create an index-to-tuple basis

    Args:
        codomain_elements: tuples representing codomain elements
    """
    idx_dict = dict(zip(range(len(codomain_elements)), codomain_elements))
    tuple_dict = dict(zip(codomain_elements, range(len(codomain_elements))))
    codomain_element_size = len(codomain_elements[0])

    def forward(i: int):
        return idx_dict[i]

    def reverse(t: Tuple[int, ...]) -> int:
        return tuple_dict[t]

    def sizes() -> Tuple[int, ...]:
        return (1, codomain_element_size)

    return Bijection(forward, reverse, sizes)
